"""Various conversion datapipes"""

from .coordinates import ConvertGeostationaryToLonLatIterDataPipe as ConvertGeostationaryToLonLat
from .coordinates import ConvertLonLatToOSGBIterDataPipe as ConvertLonLatToOSGB
from .coordinates import ConvertOSGBToLonLatIterDataPipe as ConvertOSGBToLonLat
from .numpy.gsp import ConvertGSPToNumpyIterDataPipe as ConvertGSPToNumpy
from .numpy.pv import ConvertPVToNumpyIterDataPipe as ConvertPVToNumpy
from .numpy_batch.gsp import (
    ConvertGSPToNumpyBatchIterDataPipe as ConvertGSPToNumpyBatch,
)
from .numpy_batch.nwp import (
    ConvertNWPToNumpyBatchIterDataPipe as ConvertNWPToNumpyBatch,
)
from .numpy_batch.pv import (
    ConvertPVToNumpyBatchIterDataPipe as ConvertPVToNumpyBatch,
)
from .numpy_batch.satellite import (
    ConvertSatelliteToNumpyBatchIterDataPipe as ConvertSatelliteToNumpyBatch,
)
from .numpy_batch.sensor import (
    ConvertSensorToNumpyBatchIterDataPipe as ConvertSensorToNumpyBatch,
)
from .numpy_batch.wind import (
    ConvertWindToNumpyBatchIterDataPipe as ConvertWindToNumpyBatch,
)
from .stack_xarray_to_numpy import StackXarrayIterDataPipe as StackXarray



"""Open GFS Forecast data"""

import logging

import xarray as xr

from ocf_datapipes.load.nwp.providers.utils import open_zarr_paths

_log = logging.getLogger(__name__)


def open_gfs(zarr_path) -> xr.DataArray:
    """
    Opens the GFS data

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    _log.info("Loading NWP GFS data")

    # Open data
    gfs: xr.Dataset = open_zarr_paths(zarr_path, time_dim="init_time_utc")
    nwp: xr.DataArray = gfs.to_array()

    del gfs

    nwp = nwp.rename({"variable": "channel"})
    nwp = nwp.transpose("init_time_utc", "step", "channel", "latitude", "longitude")

    return nwp



"""Convert point PV sites to image output"""

import logging
from collections import defaultdict

import numpy as np
import pandas as pd
import pvlib
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.utils import Zipper
from ocf_datapipes.utils.geospatial import (
    lon_lat_to_geostationary_area_coords,
    osgb_to_geostationary_area_coords,
    spatial_coord_type,
)

logger = logging.getLogger(__name__)


@functional_datapipe("create_pv_image")
class CreatePVImageIterDataPipe(IterDataPipe):
    """Create PV image from individual sites"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        image_datapipe: IterDataPipe,
        normalize: bool = False,
        max_num_pv_systems: int = np.inf,
        always_return_first: bool = False,
        seed: int = None,
        max_num_pv_systems_per_pixel: int = np.inf,
        normalize_by_pvlib: bool = False,
    ):
        """
        Creates a 3D data cube of PV output image x number of timesteps

        TODO Include PV System IDs or something, currently just takes the outputs

        Args:
            source_datapipe: Source datapipe of PV data
            image_datapipe: Datapipe emitting images to get the shape from, with coordinates
            normalize: Whether to normalize based off the image max, or leave raw data
            max_num_pv_systems: Max number of PV systems to consider to generate entire image
            always_return_first: Always return the first image data cube, to save computation
                Only use for if making the image at the beginning of the stack
            seed: Random seed to use if using max_num_pv_systems
            max_num_pv_systems_per_pixel: Limit the number of PV systems used when finding the
                generation timeseries in each pixel
            normalize_by_pvlib: Normalize by pvlib's poa_global based off
                tilt/orientation/capacity/lat/lon of the system

        """
        if normalize_by_pvlib and normalize:
            raise ValueError("Cannot normalize by both max, and pvlib")
        assert max_num_pv_systems_per_pixel > 0
        assert max_num_pv_systems > 0

        self.source_datapipe = source_datapipe
        self.image_datapipe = image_datapipe
        self.normalize = normalize
        self.max_num_pv_systems = max_num_pv_systems
        self.rng = np.random.default_rng(seed=seed)
        self.always_return_first = always_return_first
        self.max_num_pv_systems_per_pixel = max_num_pv_systems_per_pixel
        self.normalize_by_pvlib = normalize_by_pvlib

    def __iter__(self) -> xr.DataArray:
        for pv_systems_xr, image_xr in Zipper(self.source_datapipe, self.image_datapipe):
            pv_coords, pv_x_dim, pv_y_dim = spatial_coord_type(pv_systems_xr)
            image_coords, image_x_dim, image_y_dim = spatial_coord_type(image_xr)

            # Randomly sample systems if too many
            if self.max_num_pv_systems <= len(pv_systems_xr.pv_system_id):
                subset_of_pv_system_ids = self.rng.choice(
                    pv_systems_xr.pv_system_id,
                    size=self.max_num_pv_systems,
                    replace=False,
                )
                pv_systems_xr = pv_systems_xr.sel(pv_system_id=subset_of_pv_system_ids)

            # Find and store spatial index positions of the PV systems within the target image
            pv_position_dict = defaultdict(list)

            for i, pv_system_id in enumerate(pv_systems_xr["pv_system_id"]):
                # went for isel incase there is a duplicated pv_system_id
                pv_system = pv_systems_xr.isel(pv_system_id=i)

                if pv_coords == "osgb" and image_coords == "geostationary":
                    pv_x, pv_y = osgb_to_geostationary_area_coords(
                        x=pv_system["x_osgb"].values,
                        y=pv_system["y_osgb"].values,
                        xr_data=image_xr,
                    )
                elif pv_coords == "lon_lat" and image_coords == "geostationary":
                    pv_x, pv_y = lon_lat_to_geostationary_area_coords(
                        x=pv_system["longitude"].values,
                        y=pv_system["latitude"].values,
                        xr_data=image_xr,
                    )

                else:
                    raise NotImplementedError(
                        f"PV coords of type {pv_coords} and image coords of type {image_coords}"
                    )

                # Check the PV system is within the image
                in_range = (image_xr[image_x_dim].min() < pv_x < image_xr[image_x_dim].max()) and (
                    image_xr[image_y_dim].min() < pv_y < image_xr[image_y_dim].max()
                )

                # If normalizing by pvlib we need tilt and orientation values
                is_normalizable = (not self.normalize_by_pvlib) or np.isfinite(
                    [pv_system.orientation.values, pv_system.tilt.values]
                ).all()

                if not in_range or not is_normalizable:
                    # Skip the PV system if not inside the image
                    # Skip the PV system if pvlib normalizarion requested but cannot be completed
                    continue

                x_idx = image_xr.get_index(image_x_dim).get_indexer([pv_x], method="nearest")[0]
                y_idx = image_xr.get_index(image_y_dim).get_indexer([pv_y], method="nearest")[0]

                pv_position_dict[(y_idx, x_idx)].append(pv_system)

            # Create empty image to use for the PV Systems, assumes image has x and y coordinates
            pv_image = np.zeros(
                (
                    len(pv_systems_xr["time_utc"]),
                    len(image_xr[image_y_dim]),
                    len(image_xr[image_x_dim]),
                ),
                dtype=np.float32,
            )

            for (y_idx, x_idx), system_list in pv_position_dict.items():
                # Randomly sample systems if too many per pixel
                if self.max_num_pv_systems_per_pixel < len(system_list):
                    random_inds = self.rng.choice(
                        np.arange(len(system_list)),
                        size=self.max_num_pv_systems_per_pixel,
                        replace=False,
                    )
                    system_list = [system_list[i] for i in random_inds]

                # Find average output timeseries of all systems in pixel
                avg_generation = np.zeros_like(system_list[0].values)
                for pv_system in system_list:
                    if self.normalize_by_pvlib:
                        pv_system = _normalize_by_pvlib(pv_system)
                    avg_generation += np.nan_to_num(pv_system.values)
                avg_generation /= len(system_list)
                pv_image[:, y_idx, x_idx] = avg_generation

            if self.normalize:
                max_val = np.nanmax(pv_image)
                if max_val > 0:
                    pv_image = pv_image / max_val

            # Should return xarray
            # Same coordinates as the image xarray, so can take that
            pv_image_xr = xr.DataArray(
                data=pv_image,
                coords=(
                    ("time_utc", pv_systems_xr.time_utc.values),
                    (image_y_dim, image_xr[image_y_dim].values),
                    (image_x_dim, image_xr[image_x_dim].values),
                ),
                name="pv_image",
            )
            pv_image_xr.attrs = image_xr.attrs

            if self.always_return_first:
                while True:
                    yield pv_image_xr
            yield pv_image_xr


def _normalize_by_pvlib(pv_system):
    """
    Normalize the output by pv_libs poa_global

    Args:
        pv_system: PV System in Xarray DataArray

    Returns:
        PV System in xarray DataArray, but normalized values
    """
    # TODO Add elevation
    pvlib_loc = pvlib.location.Location(
        latitude=pv_system.latitude.values, longitude=pv_system.longitude.values
    )
    times = pd.DatetimeIndex(pv_system.time_utc.values)
    solar_position = pvlib_loc.get_solarposition(times=times)
    clear_sky = pvlib_loc.get_clearsky(times)
    total_irradiance = pvlib.irradiance.get_total_irradiance(
        pv_system.tilt.values,
        pv_system.orientation.values,
        solar_zenith=solar_position["zenith"],
        solar_azimuth=solar_position["azimuth"],
        dni=clear_sky["dni"],
        dhi=clear_sky["dhi"],
        ghi=clear_sky["ghi"],
    )
    # Guess want fraction of total irradiance on panel, to get fraction to do with capacity
    fraction_clear_sky = total_irradiance["poa_global"] / (
        clear_sky["dni"] + clear_sky["dhi"] + clear_sky["ghi"]
    )
    pv_system /= pv_system.observed_capacity_wp
    pv_system *= fraction_clear_sky
    return pv_system



"""Various utilities for use in datapipes"""

from .datapipes import LengthSetterIterDataPipe as LengthSetter
from .datapipes import RepeaterIterDataPipe as Repeater
from .datapipes import UnZipperIterDataPipe as UnZipper
from .datapipes import ZipperIterDataPipe as Zipper
from .location import Location



"""Convert point PV sites to image output"""

import logging

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.utils import Zipper
from ocf_datapipes.utils.geospatial import (
    lon_lat_to_geostationary_area_coords,
    osgb_to_geostationary_area_coords,
    spatial_coord_type,
)

logger = logging.getLogger(__name__)

MAX_TILT = 90.0
MAX_ORIENTATION = 360.0


@functional_datapipe("create_pv_metadata_image")
class CreatePVMetadataImageIterDataPipe(IterDataPipe):
    """Create PV Metadata image from individual sites"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        image_datapipe: IterDataPipe,
        normalize: bool = False,
        max_num_pv_systems: int = np.inf,
        always_return_first: bool = False,
        seed: int = None,
    ):
        """
        Creates a 2D data cube  of PV Metadata output image

        The metadata is orientation + tilt

        Args:
            source_datapipe: Source datapipe of PV data
            image_datapipe: Datapipe emitting images to get the shape from, with coordinates
            normalize: Whether to normalize based off the image max, or leave raw data
            max_num_pv_systems: Max number of PV systems to consider to generate entire image
            always_return_first: Always return the first image data cube, to save computation
                Only use for if making the image at the beginning of the stack
            seed: Random seed to use if using max_num_pv_systems
        """
        assert max_num_pv_systems > 0

        self.source_datapipe = source_datapipe
        self.image_datapipe = image_datapipe
        self.normalize = normalize
        self.max_num_pv_systems = max_num_pv_systems
        self.rng = np.random.default_rng(seed=seed)
        self.always_return_first = always_return_first

    def __iter__(self) -> xr.DataArray:
        for pv_systems_xr, image_xr in Zipper(self.source_datapipe, self.image_datapipe):
            pv_coords, pv_x_dim, pv_y_dim = spatial_coord_type(pv_systems_xr)
            image_coords, image_x_dim, image_y_dim = spatial_coord_type(image_xr)

            # Randomly sample systems if too many
            if self.max_num_pv_systems <= len(pv_systems_xr.pv_system_id.values):
                subset_of_pv_system_ids = self.rng.choice(
                    pv_systems_xr.pv_system_id,
                    size=self.max_num_pv_systems,
                    replace=False,
                )
                pv_systems_xr = pv_systems_xr.sel(pv_system_id=subset_of_pv_system_ids)

            # Create empty image to use for the PV Systems, assumes image has x and y coordinates
            pv_meta_image = np.zeros(
                (2, len(image_xr[image_y_dim]), len(image_xr[image_x_dim])),
                dtype=np.float32,
            )

            for i, pv_system_id in enumerate(pv_systems_xr["pv_system_id"]):
                # went for isel incase there is a duplicated pv_system_id
                pv_system = pv_systems_xr.isel(pv_system_id=i)

                if pv_coords == "osgb" and image_coords == "geostationary":
                    pv_x, pv_y = osgb_to_geostationary_area_coords(
                        x=pv_system["x_osgb"].values,
                        y=pv_system["y_osgb"].values,
                        xr_data=image_xr,
                    )
                elif pv_coords == "lon_lat" and image_coords == "geostationary":
                    pv_x, pv_y = lon_lat_to_geostationary_area_coords(
                        x=pv_system["longitude"].values,
                        y=pv_system["latitude"].values,
                        xr_data=image_xr,
                    )

                else:
                    raise NotImplementedError(
                        f"PV coords of type {pv_coords} and image coords of type {image_coords}"
                    )

                # Check the PV system is within the image
                in_range = (image_xr[image_x_dim].min() < pv_x < image_xr[image_x_dim].max()) and (
                    image_xr[image_y_dim].min() < pv_y < image_xr[image_y_dim].max()
                )

                if not in_range:
                    # Skip the PV system if not inside the image
                    continue

                x_idx = image_xr.get_index(image_x_dim).get_indexer([pv_x], method="nearest")[0]
                y_idx = image_xr.get_index(image_y_dim).get_indexer([pv_y], method="nearest")[0]

                # TODO: should we be using an average here? Overwiting will occur as it is
                pv_meta_image[:, y_idx, x_idx] = np.array(
                    [pv_system["tilt"], pv_system["orientation"]]
                )

            if self.normalize:
                pv_meta_image = pv_meta_image / np.array([MAX_TILT, MAX_ORIENTATION])[:, None, None]

            # Should return xarray
            # Same coordinates as the image xarray, so can take that
            pv_meta_image_xr = xr.DataArray(
                data=pv_meta_image,
                coords=(
                    # TODO fix coordinate name?
                    ("time_utc", ["tilt", "orientation"]),
                    (image_y_dim, image_xr[image_y_dim].values),
                    (image_x_dim, image_xr[image_x_dim].values),
                ),
                name="pv_meta_image",
            )
            pv_meta_image_xr.attrs = image_xr.attrs

            yield pv_meta_image_xr



"""Select the t0 time for training"""

import logging

import numpy as np
import pandas as pd
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("pick_t0_times")
class PickT0TimesIterDataPipe(IterDataPipe):
    """Picks (random) t0 times from a dataset"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        return_all: bool = False,
        shuffle: bool = False,
        dim_name: str = "time_utc",
    ):
        """
        Datapipe to yield t0 times from the input data source.

        Args:
            source_datapipe: Datapipe emitting Xarray objects.
            return_all: Whether to return all t0 values, else sample with replacement. If True, the
                default behaviour to return t0 values in order - see `shuffle` parameter.
            shuffle: If `return_all` is True this sets whether the pairs are
                shuffled before being returned.
            dim_name: The time dimension name to use.
        """
        self.source_datapipe = source_datapipe
        self.return_all = return_all
        self.shuffle = shuffle
        self.dim_name = dim_name

    def _yield_random_iter(self, xr_dataset):
        """Sample t0 with replacement"""
        while True:
            t0 = np.random.choice(xr_dataset[self.dim_name].values)
            yield t0

    def _yield_all_iter(self, xr_dataset):
        """Yield all the t0s in order, and maybe with a shuffle"""
        all_t0s = np.copy(xr_dataset[self.dim_name].values)
        if self.shuffle:
            all_t0s = np.random.permutation(all_t0s)
        for t0 in all_t0s:
            yield t0

    def __iter__(self) -> pd.Timestamp:
        xr_dataset = next(iter(self.source_datapipe))

        if len(xr_dataset[self.dim_name].values) == 0:
            raise Exception("There are no values to get t0 from")

        if self.return_all:
            return self._yield_all_iter(xr_dataset)
        else:
            return self._yield_random_iter(xr_dataset)



"""Normalize data in various ways"""

import logging
from typing import Callable, Optional, Union

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("normalize")
class NormalizeIterDataPipe(IterDataPipe):
    """Normalize the data in various methods"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        mean: Optional[Union[xr.Dataset, xr.DataArray, np.ndarray]] = None,
        std: Optional[Union[xr.Dataset, xr.DataArray, np.ndarray]] = None,
        max_value: Optional[Union[int, float]] = None,
        calculate_mean_std_from_example: bool = False,
        normalize_fn: Optional[Callable] = None,
    ):
        """
        Normalize the data with either given mean/std,

        calculated from the example itself, or a function
        This is essentially a nice wrapper instead of using the Map DataPipe

        Args:
            source_datapipe: Datapipe emitting the datasets
            mean: Means
            std: Standard Deviations
            max_value: Max value for dividing the entire example by
            calculate_mean_std_from_example: Whether to calculate the
                mean/std from the input data or not
            normalize_fn: Callable function to apply to the data to normalize it
        """
        self.source_datapipe = source_datapipe
        self.mean = mean
        self.std = std
        self.max_value = max_value
        self.calculate_mean_std_from_example = calculate_mean_std_from_example
        self.normalize_fn = normalize_fn

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Normalize the data depending on the init arguments"""

        logger.debug("Normalizing data")

        for xr_data in self.source_datapipe:
            assert xr_data is not None
            if (self.mean is not None) and (self.std is not None):
                xr_data = xr_data - self.mean
                xr_data = xr_data / self.std
            elif self.max_value is not None:
                xr_data = xr_data / self.max_value
            elif self.calculate_mean_std_from_example:
                # For Topo data for example
                xr_data -= xr_data.mean().item()
                xr_data /= xr_data.std().item()
            else:
                try:
                    logger.debug(f"Normalizing by {self.normalize_fn}")
                    xr_data_un_normalized = xr_data
                    xr_data = self.normalize_fn(xr_data_un_normalized)
                except Exception as e:
                    logger.error(f"Could not run function {self.normalize_fn}")
                    logger.debug(xr_data_un_normalized)
                    raise e

            logger.debug("Normalizing data:done")
            yield xr_data



"""Up Sample Xarray datasets Datapipe"""

import logging
from typing import Optional

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

log = logging.getLogger(__name__)


@functional_datapipe("upsample")
class UpSampleIterDataPipe(IterDataPipe):
    """Up Sample Xarray dataset with Interpolate"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        y_upsample: int,
        x_upsample: int,
        x_dim_name: str = "longitude",
        y_dim_name: str = "latitude",
        keep_same_shape: bool = False,
        round_to_dp: Optional[int] = None,
    ):
        """
        Up Sample xarray dataset/dataarrays with interpolate

        Args:
            source_datapipe: Datapipe emitting Xarray dataset
            y_upsample: up sample value in the y direction
            x_upsample: Up sample value in the x direction
            x_dim_name: X dimension name
            y_dim_name: Y dimension name
            keep_same_shape: Optional to keep the same shape. Defaults to zero.
                If True, shape is trimmed around the edges.
            round_to_dp: Try to round values to this number of decimal places.
                Default is None, so no rounding is done.
        """
        self.source_datapipe = source_datapipe
        self.y_upsample = y_upsample
        self.x_upsample = x_upsample
        self.x_dim_name = x_dim_name
        self.y_dim_name = y_dim_name
        self.keep_same_shape = keep_same_shape
        self.round_to_dp = round_to_dp

    def __iter__(self):
        """Coarsen the data on the specified dimensions"""
        for xr_data in self.source_datapipe:
            log.info("Up Sampling Data")

            # get current x and y values
            current_x_dim_values = getattr(xr_data, self.x_dim_name).values
            current_y_dim_values = getattr(xr_data, self.y_dim_name).values

            # get current interval values
            current_x_interval = np.abs(current_x_dim_values[1] - current_x_dim_values[0])
            current_y_interval = np.abs(current_y_dim_values[1] - current_y_dim_values[0])

            # new intervals
            new_x_interval = current_x_interval / self.x_upsample
            new_y_interval = current_y_interval / self.y_upsample

            if self.round_to_dp is not None:
                new_x_interval = np.round(new_x_interval, self.round_to_dp)
                new_y_interval = np.round(new_y_interval, self.round_to_dp)

            if self.keep_same_shape:
                # up sample the center of the image and keep the same shape as original image

                center_x = current_x_dim_values[int(len(current_x_dim_values) / 2)]
                center_y = current_y_dim_values[int(len(current_y_dim_values) / 2)]

                new_x_min = center_x - new_x_interval * int(len(current_x_dim_values) / 2)
                new_x_max = new_x_min + new_x_interval * (len(current_x_dim_values) - 1)

                new_y_min = center_y - new_y_interval * int(len(current_y_dim_values) / 2)
                new_y_max = new_y_min + new_y_interval * (len(current_y_dim_values) - 1)

            else:
                new_x_min = min(current_x_dim_values)
                new_x_max = max(current_x_dim_values)

                new_y_min = min(current_y_dim_values)
                new_y_max = max(current_y_dim_values)

            # round to decimals places
            if self.round_to_dp is not None:
                new_x_min = np.round(new_x_min, self.round_to_dp)
                new_x_max = np.round(new_x_max, self.round_to_dp)

                new_y_min = np.round(new_y_min, self.round_to_dp)
                new_y_max = np.round(new_y_max, self.round_to_dp)

            # get new x values
            new_x_dim_values = list(
                np.arange(
                    new_x_min,
                    new_x_max + new_x_interval,
                    new_x_interval,
                )
            )

            # get new y values
            new_y_dim_values = list(
                np.arange(
                    new_y_min,
                    new_y_max + new_y_interval,
                    new_y_interval,
                )
            )

            # check the order
            if current_x_dim_values[0] > current_x_dim_values[1]:
                new_x_dim_values = new_x_dim_values[::-1]
                new_x_max, new_x_min = new_x_min, new_x_max
                log.debug("X dims are reversed")
            if current_y_dim_values[0] > current_y_dim_values[1]:
                new_y_dim_values = new_y_dim_values[::-1]
                new_y_max, new_y_min = new_y_min, new_y_max
                log.debug("Y dims are reversed")

            log.info(
                f"Up Sampling X from ({min(current_x_dim_values)}, {current_x_interval}, "
                f"{max(current_x_dim_values)}) to ({new_x_min}, {new_x_interval}, {new_x_max})"
            )
            log.info(
                f"Up Sampling Y from ({min(current_y_dim_values)}, {current_y_interval}, "
                f"{max(current_y_dim_values)}) to ({new_y_min}, {new_y_interval}, {new_y_max})"
            )

            # resample
            xr_data = xr_data.interp(**{self.x_dim_name: new_x_dim_values})
            xr_data = xr_data.interp(**{self.y_dim_name: new_y_dim_values})

            yield xr_data



"""Create the training/validation datapipe for UK PVNet batches for all GSPs

The main public functions are:

    [1] `pvnet_all_gsp_datapipe()`
        This constructs a datapipe yielding batches with inputs for all 317 UK GSPs for random t0
        times.

    [2] `construct_sliced_data_pipeline()`
        Given a datapipe yielding t0 times, this function constructs a datapipe yielding batches
        with inputs for all 317 UK GSPs for the yielded t0 times. This function is used inside [1].

"""

import logging
from datetime import datetime
from typing import List, Optional, Tuple, Union

import xarray as xr
from torch.utils.data.datapipes._decorator import functional_datapipe
from torch.utils.data.datapipes.datapipe import IterDataPipe

from ocf_datapipes.batch import MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.batch.merge_numpy_examples_to_batch import stack_np_examples_into_batch
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.convert.numpy_batch import (
    convert_gsp_to_numpy_batch,
    convert_nwp_to_numpy_batch,
    convert_pv_to_numpy_batch,
    convert_satellite_to_numpy_batch,
)
from ocf_datapipes.load.gsp.utils import GSPLocationLookup
from ocf_datapipes.select.select_spatial_slice import (
    select_spatial_slice_meters,
    select_spatial_slice_pixels,
)
from ocf_datapipes.training.common import (
    _get_datapipes_dict,
    check_nans_in_satellite_data,
    concat_xr_time_utc,
    create_valid_t0_periods_datapipe,
    fill_nans_in_arrays,
    fill_nans_in_pv,
    normalize_gsp,
    normalize_pv,
    slice_datapipes_by_time,
)
from ocf_datapipes.utils.consts import (
    NWP_MEANS,
    NWP_STDS,
    RSS_MEAN,
    RSS_STD,
)
from ocf_datapipes.utils.location import Location

xr.set_options(keep_attrs=True)
logger = logging.getLogger("pvnet_all_gsp_datapipe")


# ---------------------------------- Utility datapipes ---------------------------------


def xr_compute(xr_data):
    """Compute the xarray object"""
    return xr_data.compute()


class SampleRepeat:
    """Use a single input element to create a list of identical values"""

    def __init__(self, num_repeats):
        """Use a single input element to create a list of identical values

        Args:
            num_repeats: Length of the returned list of duplicated values
        """
        self.num_repeats = num_repeats

    def __call__(self, x):
        """Repeat the input a number of times as a list"""
        return [x for _ in range(self.num_repeats)]


@functional_datapipe("list_map")
class ListMap(IterDataPipe):
    """Datapipe used to appky function to each item in yielded list"""

    def __init__(self, source_datapipe: IterDataPipe, func, *args, **kwargs):
        """Datapipe used to appky function to each item in yielded list.

        Args:
            source_datapipe: The source datapipe yielding lists of samples
            func: The function to apply to all items in the list
            *args: Args to pass to the function
            **kwargs: Keyword arguments to pass to the function

        """

        self.source_datapipe = source_datapipe
        self.func = func
        self._args = args
        self._kwargs = kwargs

    def __iter__(self):
        for element_list in self.source_datapipe:
            yield [self.func(x, *self._args, **self._kwargs) for x in element_list]


# ------------------------------ Multi-location datapipes ------------------------------
# These are datapipes rewritten to run on all GSPs


@functional_datapipe("select_all_gsp_spatial_slices_pixels")
class SelectAllGSPSpatialSlicePixelsIterDataPipe(IterDataPipe):
    """Select all the spatial slices"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        locations: List[Location],
        roi_height_pixels: int,
        roi_width_pixels: int,
        allow_partial_slice: bool = False,
        location_idx_name: Optional[str] = None,
    ):
        """
        Select spatial slices for all GSPs

        If `allow_partial_slice` is set to True, then slices may be made which intersect the border
        of the input data. The additional x and y cordinates that would be required for this slice
        are extrapolated based on the average spacing of these coordinates in the input data.
        However, currently slices cannot be made where the centre of the window is outside of the
        input data.

        Args:
            source_datapipe: Datapipe of Xarray data
            locations: List of all locations to create samples for
            roi_height_pixels: ROI height in pixels
            roi_width_pixels: ROI width in pixels
            allow_partial_slice: Whether to allow a partial slice.
            location_idx_name: Name for location index of unstructured grid data,
                None if not relevant
        """
        self.source_datapipe = source_datapipe
        self.locations = locations
        self.roi_height_pixels = roi_height_pixels
        self.roi_width_pixels = roi_width_pixels
        self.allow_partial_slice = allow_partial_slice
        self.location_idx_name = location_idx_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data in self.source_datapipe:
            loc_slices = []

            for location in self.locations:
                selected = select_spatial_slice_pixels(
                    xr_data,
                    location,
                    self.roi_width_pixels,
                    self.roi_height_pixels,
                    self.allow_partial_slice,
                    self.location_idx_name,
                )

                loc_slices.append(selected)

            yield loc_slices


@functional_datapipe("select_all_gsp_spatial_slice_meters")
class SelectAllGSPSpatialSliceMetersIterDataPipe(IterDataPipe):
    """Select spatial slice based off meters from point of interest"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        locations: List[Location],
        roi_height_meters: int,
        roi_width_meters: int,
        dim_name: Optional[str] = None,
    ):
        """
        Select spatial slice based off pixels from point of interest

        Args:
            source_datapipe: Datapipe of Xarray data
            locations: List of all locations to create samples for
            roi_width_meters: ROI width in meters
            roi_height_meters: ROI height in meters
            dim_name: Dimension name to select for ID, None for coordinates

        Notes:
            Using spatial slicing based on distance rather than number of pixels will often yield
            slices which can vary by 1 pixel in height and/or width.

            E.g. Suppose the Xarray data has x-coords = [1,2,3,4,5]. We want to slice a spatial
            window with a size which equates to 2.2 along the x-axis. If we choose to slice around
            the point x=3 this will slice out the x-coords [2,3,4]. If we choose to slice around the
            point x=2.5 this will slice out the x-coords [2,3]. Hence the returned slice can have
            size either 2 or 3 in the x-axis depending on the spatial location selected.

            Also, if selecting over a large span of latitudes, this may also causes pixel sizes of
            the yielded outputs to change. For example, if the Xarray data is on a regularly spaced
            longitude-latitude grid, then the structure of the grid means that the longitudes near
            to the poles are spaced closer together (measured in meters) than at the equator. So
            slices near the equator will have less pixels in the x-axis than slices taken near the
            poles.
        """
        self.source_datapipe = source_datapipe
        self.locations = locations
        self.roi_width_meters = roi_width_meters
        self.roi_height_meters = roi_height_meters
        self.dim_name = dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data in self.source_datapipe:
            loc_slices = []

            for location in self.locations:
                selected = select_spatial_slice_meters(
                    xr_data=xr_data,
                    location=location,
                    roi_width_meters=self.roi_width_meters,
                    roi_height_meters=self.roi_height_meters,
                    dim_name=self.dim_name,
                )

                loc_slices.append(selected)

            yield loc_slices


# ------------------------------- Time pipeline functions ------------------------------


def construct_time_pipeline(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> Tuple[IterDataPipe, IterDataPipe]:
    """Construct time pipeline for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time for time datapipe.
        end_time: Maximum time for time datapipe.
    """

    datapipes_dict = _get_datapipes_dict(config_filename)

    # Get config
    config = datapipes_dict.pop("config")

    if (start_time is not None) or (end_time is not None):
        datapipes_dict["gsp"] = datapipes_dict["gsp"].filter_times(start_time, end_time)

    # Get overlapping time periods
    t0_datapipe = create_t0_datapipe(
        datapipes_dict,
        configuration=config,
        shuffle=True,
    )

    return t0_datapipe


def create_t0_datapipe(
    datapipes_dict: dict,
    configuration: Configuration,
    shuffle: bool = True,
):
    """
    Takes source datapipes and returns datapipes of appropriate t0 times.

    The t0 times are sampled without replacement.

    Args:
        datapipes_dict: Dictionary of datapipes of input sources for which we want to select
            appropriate t0 times.
        configuration: Configuration object for inputs.
        shuffle: Whether to use the internal shuffle function when yielding times. Else
            location times will be heavily ordered.

    Returns:
        t0 datapipe

    """
    valid_t0_periods_datapipe = create_valid_t0_periods_datapipe(
        datapipes_dict,
        configuration,
        key_for_t0="gsp",
    )

    t0_datapipe = valid_t0_periods_datapipe.pick_t0_times(return_all=True, shuffle=shuffle)

    return t0_datapipe


# ------------------------------- Space pipeline functions -----------------------------


def slice_datapipes_by_space_all_gsps(
    datapipes_dict: dict,
    locations: list[Location],
    configuration: Configuration,
) -> None:
    """Slice the dictionary of datapipes by space in-place"""
    conf_nwp = configuration.input_data.nwp
    conf_sat = configuration.input_data.satellite

    if "nwp" in datapipes_dict:
        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            datapipes_dict["nwp"][nwp_key] = nwp_datapipe.select_all_gsp_spatial_slices_pixels(
                locations,
                roi_width_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_width,
                roi_height_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_height,
            )

    if "sat" in datapipes_dict:
        datapipes_dict["sat"] = datapipes_dict["sat"].select_all_gsp_spatial_slices_pixels(
            locations,
            roi_width_pixels=conf_sat.satellite_image_size_pixels_width,
            roi_height_pixels=conf_sat.satellite_image_size_pixels_height,
        )

    if "pv" in datapipes_dict:
        # No spatial slice for PV since it is always the same, just repeat for GSPs
        datapipes_dict["pv"] = datapipes_dict["pv"].map(SampleRepeat(len(locations)))

    # GSP always assumed to be in data
    datapipes_dict["gsp"] = datapipes_dict["gsp"].select_all_gsp_spatial_slice_meters(
        locations,
        roi_width_meters=1,
        roi_height_meters=1,
        dim_name="gsp_id",
    )


# -------------------------------- Processing functions --------------------------------


def pre_spatial_slice_process(datapipes_dict, configuration):
    """Apply pre-processing steps to the dictionary of datapipes in place

    These steps are normalisation and recombining past and future GSP/PV data
    """
    conf_nwp = configuration.input_data.nwp

    if "nwp" in datapipes_dict:
        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            datapipes_dict["nwp"][nwp_key] = nwp_datapipe.map(xr_compute).normalize(
                mean=NWP_MEANS[conf_nwp[nwp_key].nwp_provider],
                std=NWP_STDS[conf_nwp[nwp_key].nwp_provider],
            )

    if "sat" in datapipes_dict:
        datapipes_dict["sat"] = (
            datapipes_dict["sat"].map(xr_compute).normalize(mean=RSS_MEAN, std=RSS_STD)
        )

    if "pv" in datapipes_dict:
        # Recombine PV arrays - see function doc for further explanation
        datapipes_dict["pv"] = (
            datapipes_dict["pv"]
            .zip_ocf(datapipes_dict["pv_future"])
            .map(concat_xr_time_utc)
            .map(normalize_pv)
            .map(fill_nans_in_pv)
        )

        del datapipes_dict["pv_future"]

    # GSP always assumed to be in data
    # Recombine GSP arrays - see function doc for further explanation
    datapipes_dict["gsp"] = (
        datapipes_dict["gsp"]
        .zip_ocf(datapipes_dict["gsp_future"])
        .map(concat_xr_time_utc)
        .map(normalize_gsp)
    )

    del datapipes_dict["gsp_future"]


def post_spatial_slice_process(datapipes_dict, check_satellite_no_nans=False):
    """Convert the dictionary of datapipes to NumpyBatches, combine, and fill nans"""

    numpy_modalities = []

    if "nwp" in datapipes_dict:
        nwp_numpy_modalities = dict()

        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            nwp_numpy_modalities[nwp_key] = nwp_datapipe.list_map(convert_nwp_to_numpy_batch).map(
                stack_np_examples_into_batch
            )

        # Combine the NWPs into NumpyBatch
        nwp_numpy_modalities = MergeNWPNumpyModalities(nwp_numpy_modalities)
        numpy_modalities.append(nwp_numpy_modalities)

    if "sat" in datapipes_dict:
        numpy_modalities.append(
            datapipes_dict["sat"]
            .list_map(convert_satellite_to_numpy_batch)
            .map(stack_np_examples_into_batch)
        )

    if "pv" in datapipes_dict:
        numpy_modalities.append(
            datapipes_dict["pv"]
            .list_map(convert_pv_to_numpy_batch)
            .map(stack_np_examples_into_batch)
        )

    # GSP always assumed to be in data
    numpy_modalities.append(
        datapipes_dict["gsp"].list_map(convert_gsp_to_numpy_batch).map(stack_np_examples_into_batch)
    )

    # Combine all the data sources
    combined_datapipe = MergeNumpyModalities(numpy_modalities).add_sun_position(modality_name="gsp")

    if check_satellite_no_nans:
        # in production we don't want any nans in the satellite data
        combined_datapipe = combined_datapipe.map(check_nans_in_satellite_data)

    combined_datapipe = combined_datapipe.map(fill_nans_in_arrays)

    return combined_datapipe


# --------------------------- High level pipeline functions ----------------------------


def construct_sliced_data_pipeline(
    config_filename: str,
    t0_datapipe: IterDataPipe,
    production: bool = False,
    check_satellite_no_nans: bool = False,
) -> IterDataPipe:
    """Constructs data pipeline for the input data config file.

    This yields samples from the location and time datapipes.

    Args:
        config_filename: Path to config file.
        t0_datapipe: Datapipe yielding times.
        production: Whether constucting pipeline for production inference.
        check_satellite_no_nans: Whether to check that satellite data has no nans.
    """

    datapipes_dict = _get_datapipes_dict(
        config_filename,
        production=production,
    )

    # Get the location objects for all 317 regional GSPs
    gsp_id_to_loc = GSPLocationLookup()
    locations = [gsp_id_to_loc(gsp_id) for gsp_id in range(1, 318)]

    # Pop config
    configuration = datapipes_dict.pop("config")

    # Slice all of the datasets by time - this is an in-place operation
    slice_datapipes_by_time(datapipes_dict, t0_datapipe, configuration, production)

    # Run compute and normalise all the data
    pre_spatial_slice_process(datapipes_dict, configuration)

    # Slice all of the datasets by space - this is an in-place operation
    slice_datapipes_by_space_all_gsps(datapipes_dict, locations, configuration)

    # Convert to NumpyBatch
    combined_datapipe = post_spatial_slice_process(datapipes_dict)

    return combined_datapipe


def pvnet_all_gsp_datapipe(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> IterDataPipe:
    """
    Construct pvnet pipeline for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time at which a sample can be selected.
        end_time: Maximum time at which a sample can be selected.
    """

    # Open datasets from the config and filter to useable times
    t0_datapipe = construct_time_pipeline(
        config_filename,
        start_time,
        end_time,
    )

    # Shard after we have the times. These are already shuffled so no need to shuffle again
    t0_datapipe = t0_datapipe.sharding_filter()

    # In this function we re-open the datasets to make a clean separation before/after sharding
    # This function
    datapipe = construct_sliced_data_pipeline(
        config_filename,
        t0_datapipe,
    )

    return datapipe



import numpy as np
from ocf_datapipes.utils.utils import searchsorted
from ocf_datapipes.utils.utils import combine_to_single_dataset, uncombine_from_single_dataset
import xarray as xr
import pytest


@pytest.fixture()
def xarray_dict_sample(pv_xarray_data, nwp_gfs_data):
    # Use already created data pieces to make batch
    da_pv = pv_xarray_data.rename(dict(datetime="time_utc"))
    da_nwp = nwp_gfs_data.rename(dict(time="init_time_utc")).to_array()

    xarray_sample = {
        "pv": da_pv.isel(time_utc=slice(0, 2)),
        "nwp": da_nwp.isel(init_time_utc=slice(1, 2)),
    }
    return xarray_sample


def test_searchsorted():
    ys = np.array([1, 2, 3, 4, 5], dtype=np.float32)
    assert searchsorted(ys, 2.1) == 2
    ys_r = np.array([5, 4, 3, 2, 1], dtype=np.float32)
    assert searchsorted(ys_r, 2.1, assume_ascending=False) == 3


def test_combine_to_single_dataset(xarray_dict_sample):
    ds_comb = combine_to_single_dataset(xarray_dict_sample)
    # Expected data type
    assert isinstance(ds_comb, xr.Dataset)
    # Expected data variables
    assert set(ds_comb.keys()) == set(xarray_dict_sample.keys())


def test_combine_to_single_dataset_same_init_time(xarray_dict_sample):

    # set nwp init_time_utc to be the same
    xarray_dict_sample["nwp"] = xarray_dict_sample["nwp"].isel(init_time_utc=0)

    # get a vector of init times
    init_time_utcs = [xarray_dict_sample["nwp"].init_time_utc.values] * 11

    # need to have the dims
    # - target_time_utc
    # - channel
    # - latitude
    # - longitude
    # with coordinates of step and init_time_utc
    # rename step to target_time_utc, and add init_time_utc and step
    xarray_dict_sample["nwp"] = xarray_dict_sample["nwp"].rename({"variable": "channel"})
    xarray_dict_sample["nwp"] = xarray_dict_sample["nwp"].rename({"step": "target_time_utc"})
    xarray_dict_sample["nwp"] = xarray_dict_sample["nwp"].assign_coords(
        {"init_time_utc": ("target_time_utc", init_time_utcs)}
    )
    xarray_dict_sample["nwp"] = xarray_dict_sample["nwp"].assign_coords(
        {"step": ("target_time_utc", range(len(init_time_utcs)))}
    )
    # order by target_time_utc, channel, latitude, longitude
    xarray_dict_sample["nwp"] = xarray_dict_sample["nwp"].transpose(
        "target_time_utc", "channel", "latitude", "longitude"
    )

    ds_comb = combine_to_single_dataset(xarray_dict_sample)

    # Expected data type
    assert isinstance(ds_comb, xr.Dataset)
    assert len(ds_comb["nwp__init_time_utc"].values) == 11
    # Expected data variables
    assert set(ds_comb.keys()) == set(xarray_dict_sample.keys())


def test_uncombine_from_single_dataset(xarray_dict_sample):
    ds_comb = combine_to_single_dataset(xarray_dict_sample)

    recompiled_xarray_dict_sample = uncombine_from_single_dataset(ds_comb)

    # Right type
    assert isinstance(recompiled_xarray_dict_sample, type(xarray_dict_sample))

    # Keys are the same
    assert set(recompiled_xarray_dict_sample.keys()) == set(xarray_dict_sample.keys())

    # xarray object under each key is the same
    for k in xarray_dict_sample.keys():
        # Right type
        assert isinstance(recompiled_xarray_dict_sample[k], type(xarray_dict_sample[k]))

        # Coord values are the same
        for coord in xarray_dict_sample[k].coords:
            coord_values_same = (
                recompiled_xarray_dict_sample[k][coord].values
                == xarray_dict_sample[k][coord].values
            ).all()
            assert coord_values_same, f"Coord values under key: {coord} are different"

        # Values are the same
        np.testing.assert_array_equal(
            recompiled_xarray_dict_sample[k].values, xarray_dict_sample[k].values
        )



import numpy as np

from ocf_datapipes.utils.utils import _trig_transform, trigonometric_datetime_transformation


def test_trig_transform():
    input = np.array([0, 0.5, 1]) * np.pi
    expected_sine = np.array([0.0, 1.0, 0.0])
    expected_cosine = np.array([1.0, 0.0, -1.0])
    actual_sine, actual_cosine = _trig_transform(input, 2 * np.pi)
    assert np.allclose(expected_cosine, actual_cosine) and np.allclose(expected_sine, actual_sine)


def test_trigonometric_datetime_transformations():
    input = np.array(
        [
            "2014-01-01T00:00:00.000000000",
            "2014-01-01T01:00:00.000000000",
            "2014-01-02T00:00:00.000000000",
            "2014-02-01T00:00:00.000000000",
        ],
        dtype="datetime64[ns]",
    )

    expected_output = np.array(
        [
            [
                np.sin(2 * np.pi * 1 / 12),
                np.cos(2 * np.pi * 1 / 12),
                np.sin(2 * np.pi * 1 / 366),
                np.cos(2 * np.pi * 1 / 366),
                np.sin(2 * np.pi * 0 / 24),
                np.cos(2 * np.pi * 0 / 24),
            ],
            [
                np.sin(2 * np.pi * 1 / 12),
                np.cos(2 * np.pi * 1 / 12),
                np.sin(2 * np.pi * 1 / 366),
                np.cos(2 * np.pi * 1 / 366),
                np.sin(2 * np.pi * 1 / 24),
                np.cos(2 * np.pi * 1 / 24),
            ],
            [
                np.sin(2 * np.pi * 1 / 12),
                np.cos(2 * np.pi * 1 / 12),
                np.sin(2 * np.pi * 2 / 366),
                np.cos(2 * np.pi * 2 / 366),
                np.sin(2 * np.pi * 0 / 24),
                np.cos(2 * np.pi * 0 / 24),
            ],
            [
                np.sin(2 * np.pi * 2 / 12),
                np.cos(2 * np.pi * 2 / 12),
                np.sin(2 * np.pi * 1 / 366),
                np.cos(2 * np.pi * 1 / 366),
                np.sin(2 * np.pi * 0 / 24),
                np.cos(2 * np.pi * 0 / 24),
            ],
        ]
    )

    output = trigonometric_datetime_transformation(input)
    assert np.allclose(output, expected_output)
    assert np.allclose(np.sum(output**2, axis=1), 3.0 * np.ones(output.shape[0]))



"""Datapipe and utils to load PV data from NetCDF for training"""

import io
import logging
from pathlib import Path
from typing import Union

import fsspec
import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.config.model import Wind
from ocf_datapipes.load.wind.utils import (
    put_wind_data_into_an_xr_dataarray,
)

_log = logging.getLogger(__name__)


@functional_datapipe("open_wind_netcdf")
class OpenWindFromNetCDFIterDataPipe(IterDataPipe):
    """Datapipe to load NetCDF"""

    def __init__(
        self,
        wind: Wind,
    ):
        """Datapipe to load Wind from NetCDF.

        Args:
            wind: wind configuration
        """
        super().__init__()
        self.wind = wind
        self.wind_power_filenames = [
            wind_files_group.wind_filename for wind_files_group in wind.wind_files_groups
        ]
        self.wind_metadata_filenames = [
            wind_files_group.wind_metadata_filename for wind_files_group in wind.wind_files_groups
        ]

    def __iter__(self):
        wind_array_list = []
        for i in range(len(self.wind_power_filenames)):
            wind_array: xr.DataArray = load_everything_into_ram(
                self.wind_power_filenames[i],
                self.wind_metadata_filenames[i],
            )
            wind_array_list.append(wind_array)

        wind_array = xr.concat(wind_array_list, dim="wind_system_id")

        while True:
            yield wind_array


def _load_wind_metadata(filename: str) -> pd.DataFrame:
    """Return pd.DataFrame of PV metadata.

    Shape of the returned pd.DataFrame for Passiv PV data:
        Index: system_id
        Columns: latitude, longitude, ml_id
    """
    _log.info(f"Loading Wind metadata from {filename}")

    index_col = "system_id"
    df_metadata = pd.read_csv(filename, index_col=index_col)

    # Drop if exists
    df_metadata.drop(columns="Unnamed: 0", inplace=True, errors="ignore")

    # Add ml_id column if not in metadata already
    if "ml_id" not in df_metadata.columns:
        df_metadata["ml_id"] = np.nan

    _log.info(f"Found {len(df_metadata)} Wind systems in {filename}")

    return df_metadata


def load_everything_into_ram(
    generation_filename,
    metadata_filename,
    estimated_capacity_percentile: float = 100,
) -> xr.DataArray:
    """Load Wind data into xarray DataArray in RAM.

    Args:
        generation_filename: Filepath to the Wind generation data
        metadata_filename: Filepath to the Wind metadata
        estimated_capacity_percentile: Percentile used as the estimated capacity for each PV
            system. Recommended range is 99-100.
    """

    # load metadata
    df_metadata = _load_wind_metadata(metadata_filename)
    # Load pd.DataFrame of power and pd.Series of capacities:
    df_gen, estimated_capacities = _load_wind_generation_and_capacity(
        generation_filename,
        estimated_capacity_percentile=estimated_capacity_percentile,
    )
    # Drop systems where all values are NaN
    df_gen.dropna(axis="columns", how="all", inplace=True)
    estimated_capacities = estimated_capacities[df_gen.columns]

    # Ensure systems are consistant between generation data, and metadata
    common_systems = list(np.intersect1d(df_metadata.index, df_gen.columns))
    df_gen = df_gen[common_systems]
    df_metadata = df_metadata.loc[common_systems]
    estimated_capacities = estimated_capacities.loc[common_systems]
    # Compile data into an xarray DataArray
    xr_array = put_wind_data_into_an_xr_dataarray(
        df_gen=df_gen,
        observed_system_capacities=estimated_capacities,
        nominal_system_capacities=df_metadata.capacity_megawatts,
        ml_id=df_metadata.ml_id,
        latitude=df_metadata.latitude,
        longitude=df_metadata.longitude,
    )
    # Sanity checks
    time_utc = pd.DatetimeIndex(xr_array.time_utc)
    assert time_utc.is_monotonic_increasing
    assert time_utc.is_unique

    return xr_array


def _load_wind_generation_and_capacity(
    filename: Union[str, Path],
    estimated_capacity_percentile: float = 99,
) -> tuple[pd.DataFrame, pd.Series]:
    """Load the PV data and estimates the capacity for each PV system.

    The capacity is estimated by taking the max value across all datetimes in the input file.

    Args:
        filename: The filename (netcdf) of the wind data to load
        estimated_capacity_percentile: Percentile used as the estimated capacity for each PV
            system. Recommended range is 99-100.

    Returns:
        DataFrame of power output in watts. Columns are PV systems, rows are datetimes
        Series of PV system estimated capacities in watts
    """

    _log.info(f"Loading wind power data from {filename}.")

    with fsspec.open(filename, mode="rb") as file:
        file_bytes = file.read()

    _log.info("Loaded wind power bytes, now converting to xarray")
    with io.BytesIO(file_bytes) as file:
        ds_gen = xr.load_dataset(file, engine="h5netcdf")

    df_gen = ds_gen.to_dataframe()
    # Fix data types
    df_gen = df_gen.astype(np.float32)
    df_gen.columns = df_gen.columns.astype(np.int64)

    _log.info("Loaded wind PV power data and converting to pandas.")
    estimated_capacities = df_gen.quantile(estimated_capacity_percentile / 100)

    # Remove systems with no generation data
    mask = estimated_capacities > 0
    estimated_capacities = estimated_capacities[mask]
    df_gen = df_gen.loc[:, mask]
    _log.info(
        "After loading:"
        f" {len(df_gen)} wind power datetimes."
        f" {len(df_gen.columns)} wind power wind system IDs."
    )

    # Sanity checks
    assert not df_gen.columns.duplicated().any()
    assert not df_gen.index.duplicated().any()
    assert np.isfinite(estimated_capacities).all()
    assert (estimated_capacities > 0).all()
    assert np.array_equal(df_gen.columns, estimated_capacities.index)

    return df_gen, estimated_capacities



"""location"""

from typing import Optional

import numpy as np
from pydantic import BaseModel, Field, validator


class Location(BaseModel):
    """Represent a spatial location."""

    coordinate_system: Optional[str] = "osgb"  # ["osgb", "lon_lat", "geostationary", "idx"]
    x: float
    y: float
    id: Optional[int] = Field(None)

    @validator("coordinate_system", pre=True, always=True)
    def validate_coordinate_system(cls, v):
        """Validate 'coordinate_system'"""
        allowed_coordinate_systen = ["osgb", "lon_lat", "geostationary", "idx"]
        if v not in allowed_coordinate_systen:
            raise ValueError(f"coordinate_system = {v} is not in {allowed_coordinate_systen}")
        return v

    @validator("x")
    def validate_x(cls, v, values):
        """Validate 'x'"""
        min_x: float
        max_x: float
        if "coordinate_system" not in values:
            raise ValueError("coordinate_system is incorrect")
        co = values["coordinate_system"]
        if co == "osgb":
            min_x, max_x = -103976.3, 652897.98
        if co == "lon_lat":
            min_x, max_x = -180, 180
        if co == "geostationary":
            min_x, max_x = -5568748.275756836, 5567248.074173927
        if co == "idx":
            min_x, max_x = 0, np.inf
        if v < min_x or v > max_x:
            raise ValueError(f"x = {v} must be within {[min_x, max_x]} for {co} coordinate system")
        return v

    @validator("y")
    def validate_y(cls, v, values):
        """Validate 'y'"""
        min_y: float
        max_y: float
        if "coordinate_system" not in values:
            raise ValueError("coordinate_system is incorrect")
        co = values["coordinate_system"]
        if co == "osgb":
            min_y, max_y = -16703.87, 1199851.44
        if co == "lon_lat":
            min_y, max_y = -90, 90
        if co == "geostationary":
            min_y, max_y = 1393687.2151494026, 5570748.323202133
        if co == "idx":
            min_y, max_y = 0, np.inf
        if v < min_y or v > max_y:
            raise ValueError(f"y = {v} must be within {[min_y, max_y]} for {co} coordinate system")
        return v



""" Make fake PV data """

from datetime import datetime

import numpy as np

from ocf_datapipes.batch import BatchKey
from ocf_datapipes.batch.fake.utils import get_n_time_steps_from_config, make_time_utc
from ocf_datapipes.config.model import Configuration


def make_fake_pv_data(
    configuration: Configuration, t0_datetime_utc: datetime, batch_size: int = 8
) -> dict:
    """
    Make Fake PV data ready for ML model

    Args:
        configuration: configuration object
        t0_datetime_utc: one datetime for when t0 is
        batch_size: Integer batch size to create

    Returns: dictionary of pv items
    """

    pv_config = configuration.input_data.pv
    if pv_config is None:
        return {}

    n_pv_systems = pv_config.n_pv_systems_per_example
    n_fourier_features = 8

    time_utc = make_time_utc(
        batch_size=batch_size,
        history_minutes=pv_config.history_minutes,
        forecast_minutes=pv_config.forecast_minutes,
        t0_datetime_utc=t0_datetime_utc,
        time_resolution_minutes=pv_config.time_resolution_minutes,
    )
    n_times = time_utc.shape[1]

    batch = {}
    batch[BatchKey.pv_time_utc] = time_utc  # Seconds since UNIX epoch (1970-01-01).
    batch[BatchKey.pv] = np.random.random((batch_size, n_times, n_pv_systems))
    batch[BatchKey.pv_t0_idx] = get_n_time_steps_from_config(
        input_data_configuration=pv_config, include_forecast=False
    )
    batch[BatchKey.pv_ml_id] = np.random.randint(0, 100, (batch_size, n_pv_systems))
    batch[BatchKey.pv_id] = np.random.randint(0, 1000, (batch_size, n_pv_systems))
    batch[BatchKey.pv_mask] = np.random.randint(0, 1, (batch_size, n_pv_systems))
    batch[BatchKey.pv_latitude] = np.random.randint(0, 10**6, (batch_size, n_pv_systems))
    batch[BatchKey.pv_longitude] = np.random.randint(0, 10**6, (batch_size, n_pv_systems))

    batch[BatchKey.pv_latitude_fourier] = np.random.random(
        (batch_size, n_pv_systems, n_fourier_features)
    )
    batch[BatchKey.pv_longitude_fourier] = np.random.random(
        (batch_size, n_pv_systems, n_fourier_features)
    )
    batch[BatchKey.pv_time_utc_fourier] = np.random.random(
        (batch_size, n_times, n_fourier_features)
    )

    return batch



"""Convert NWP to NumpyBatch"""

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import NWPBatchKey, NWPNumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float


def convert_nwp_to_numpy_batch(xr_data):
    """Convert from Xarray to NWPBatchKey"""

    example: NWPNumpyBatch = {
        NWPBatchKey.nwp: xr_data.values,
        NWPBatchKey.nwp_t0_idx: xr_data.attrs["t0_idx"],
        NWPBatchKey.nwp_channel_names: xr_data.channel.values,
        NWPBatchKey.nwp_init_time_utc: datetime64_to_float(xr_data.init_time_utc.values),
        NWPBatchKey.nwp_step: (xr_data.step.values / np.timedelta64(1, "h")).astype(np.int64),
    }

    if "target_time_utc" in xr_data.coords:
        target_time = xr_data.target_time_utc.values
        example[NWPBatchKey.nwp_target_time_utc] = datetime64_to_float(target_time)

    for batch_key, dataset_key in (
        (NWPBatchKey.nwp_y_osgb, "y_osgb"),
        (NWPBatchKey.nwp_x_osgb, "x_osgb"),
    ):
        if dataset_key in xr_data.coords:
            example[batch_key] = xr_data[dataset_key].values

    return example


@functional_datapipe("convert_nwp_to_numpy_batch")
class ConvertNWPToNumpyBatchIterDataPipe(IterDataPipe):
    """Convert NWP Xarray objects to NWPNumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert NWP Xarray objecs to NWPNumpyBatch ones

        Args:
            source_datapipe: Datapipe emitting NWP Xarray objects
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NWPNumpyBatch:
        """Convert from Xarray to NWPBatchKey"""
        for xr_data in self.source_datapipe:
            yield convert_nwp_to_numpy_batch(xr_data)



"""Loading sensor data (i.e. weather station, rainfall radar, etc.)"""

from .awos import OpenAWOSFromNetCDFIterDataPipe as OpenAWOSFromNetCDF
from .meteomatics import OpenMeteomaticsFromZarrIterDataPipe as OpenMeteomaticsFromZarr



"""Wrapper for Power Perceiver Production Data Pipeline"""

from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray
from torch.utils.data.datapipes.datapipe import IterDataPipe

from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenPVFromNetCDF
from ocf_datapipes.training.common import normalize_pv

xarray.set_options(keep_attrs=True)

# default is set to 1000
BUFFERSIZE = 1000


def simple_pv_datapipe(
    configuration_filename: Union[Path, str],
) -> IterDataPipe:
    """Create the simple datapipe which loads PV data.

    This could be used for training an autoregressive model for site-level PV output.

    Args:
        configuration_filename: Name of the configuration

    Returns:
        IterDataPipe yielding batches of samples
    """

    # Load configuration
    configuration: Configuration = load_yaml_configuration(configuration_filename)

    # Unpack for convenience
    pv_config = configuration.input_data.pv

    # Open dataset
    pv_datapipe = OpenPVFromNetCDF(pv=pv_config)

    # Fork datapipe for raw data, time slicing, and location slicing uses
    pv_datapipe, pv_t0_datapipe, pv_time_periods_datapipe, pv_location_datapipe = pv_datapipe.fork(
        4, buffer_size=BUFFERSIZE
    )

    # Preprocess PV timeseries
    pv_datapipe = pv_datapipe.normalize(normalize_fn=normalize_pv)

    # Getting locations - Construct datapipe yielding the locations to be used for each sample
    location_datapipe = pv_location_datapipe.pick_locations()

    # Slice systems
    pv_datapipe = pv_datapipe.select_id(location_datapipe=location_datapipe, data_source_name="pv")

    # Add t0 idx
    pv_datapipe = pv_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(minutes=pv_config.time_resolution_minutes),
        history_duration=timedelta(minutes=pv_config.history_minutes),
    )

    # Get contiguous time periods
    pv_time_periods_datapipe = pv_time_periods_datapipe.find_contiguous_t0_time_periods(
        sample_period_duration=timedelta(minutes=pv_config.time_resolution_minutes),
        history_duration=timedelta(minutes=pv_config.history_minutes),
        forecast_duration=timedelta(minutes=pv_config.forecast_minutes),
    )

    # Select time periods
    pv_t0_datapipe = pv_t0_datapipe.filter_time_periods(time_periods=pv_time_periods_datapipe)
    pv_t0_datapipe = pv_t0_datapipe.pick_t0_times()

    # Take time slices
    pv_datapipe = pv_datapipe.select_time_slice(
        t0_datapipe=pv_t0_datapipe,
        history_duration=timedelta(minutes=pv_config.history_minutes),
        forecast_duration=timedelta(minutes=pv_config.forecast_minutes),
        sample_period_duration=timedelta(minutes=pv_config.time_resolution_minutes),
    )

    # Convert to numpybatch object
    pv_datapipe = pv_datapipe.convert_pv_to_numpy_batch()

    # Add solar coordinates
    pv_datapipe = pv_datapipe.add_sun_position(modality_name="pv")

    # Batch the data with batch size 4
    pv_datapipe = pv_datapipe.batch(4).merge_numpy_batch()

    return pv_datapipe



"""
This is a class function infills PV data via interpolation
"""

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("pv_interpolate_infill")
class PVInterpolateInfillIterDataPipe(IterDataPipe):
    """Infills missing values via interpolation"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        time_resolution_minutes: int = 5,
        fill_n: int = 3,
        allow_add_timestamps: bool = False,
    ):
        """Infills missing values via interpolation

        Args:
            source_datapipe: Datapipe emitting Xarray Dataset with time_utc coordinate
            time_resolution_minutes: Time resolution of output Dataset
            fill_n: Number of steps of size `time_resolution_minutes` forward and backward where
                NaNs can be infilled via interpolation.
            allow_add_timestamps: If True, new timestamps will be created in the output dataset
                which were mising in the input dataset, as long as they can be filled. Else the
                output dataset will have the same timestamps as the input dataset.
        """

        self.source_datapipe = source_datapipe
        self.time_resolution_minutes = time_resolution_minutes
        self.fill_n = fill_n
        self.allow_add_timestamps = allow_add_timestamps

    def __iter__(self) -> xr.DataArray():
        """Infills missing values via interpolation

        Returns:
            Dataset with additional values infilled
        """

        # Reading the Xarray dataset
        for ds in self.source_datapipe:
            # Resample to 5-minutely and interpolate up to 15 minutes ahead.

            # Using pandas gives us more options in the interpolation
            df_interp = (
                ds.to_pandas()
                .resample(f"{self.time_resolution_minutes}min")
                .interpolate(
                    method="time",
                    limit=self.fill_n,
                    limit_direction="both",
                )
            )
            # Create new copy of the DataArray with correct time index shape
            ds_interp = ds.sel(time_utc=df_interp.index, method="nearest")

            # Fill with interpolated values
            ds_interp.values[:] = df_interp.values

            if not self.allow_add_timestamps:
                # Filter to the same timestamps as the input
                # For the timestamps dropped here, all values are infilled values
                ds_interp = ds_interp.sel(time_utc=ds.time_utc)

            yield ds_interp



"""Selects time slice from satellite, GSP data, or other xarray objects, and masks with dropout"""

from datetime import timedelta
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("draw_dropout_time")
class DrawDropoutTimeIterDataPipe(IterDataPipe):
    """Generates dropout times. The times are absolute values, not timedeltas."""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        dropout_timedeltas: List[timedelta],
        dropout_frac: Optional[float] = 0,
    ):
        """Generates dropout times. The times are absolute values, not timedeltas.

        Args:
            source_datapipe: Datapipe of t0 times
            dropout_timedeltas: List of timedeltas. We randonly select the delay for each time from
                this list. These should be negative timedeltas w.r.t time t0.
            dropout_frac: Fraction of samples subject to dropout
        """
        self.source_datapipe = source_datapipe
        self.dropout_timedeltas = dropout_timedeltas
        self.dropout_frac = dropout_frac
        if dropout_timedeltas is not None:
            assert len(dropout_timedeltas) >= 1, "Must include list of relative dropout timedeltas"
            assert all(
                [t < timedelta(minutes=0) for t in dropout_timedeltas]
            ), "dropout timedeltas must be negative"
        assert 0 <= dropout_frac <= 1

    def __iter__(self):
        for t0 in self.source_datapipe:
            if (self.dropout_timedeltas is None) or (np.random.uniform() >= self.dropout_frac):
                dropout_time = None
            else:
                t0_datetime_utc = pd.Timestamp(t0)
                dt = np.random.choice(self.dropout_timedeltas)
                dropout_time = t0_datetime_utc + dt

            yield dropout_time


@functional_datapipe("apply_dropout_time")
class ApplyDropoutTimeIterDataPipe(IterDataPipe):
    """Masks an xarray object to replace values that come after the dropout time with NaN."""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        dropout_time_datapipe: IterDataPipe,
    ):
        """Masks an xarray object to replace values that come after the dropout time with NaN.

        Args:
            source_datapipe: Datapipe of Xarray objects
            dropout_time_datapipe: Datapipe of dropout times
        """
        self.source_datapipe = source_datapipe
        self.dropout_time_datapipe = dropout_time_datapipe

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, dropout_time in self.source_datapipe.zip_ocf(self.dropout_time_datapipe):
            if dropout_time is None:
                yield xr_data
            else:
                # This replaces the times after the dropout with NaNs
                yield xr_data.where(xr_data.time_utc <= dropout_time)



"""Utilities for parallelizing code"""

from collections import deque
from concurrent import futures
from typing import Callable, Iterator


def run_with_threadpool(
    data_iterator: Iterator,
    fn: Callable,
    max_workers: int,
    scheduled_tasks: int,
    **threadpool_kwargs
):
    """
    Copied from the ThreadPoolMapper

    Allows for running a function over data with a threadpool
    in parallel

    Args:
        data_iterator: Iterator of data to run the function over
        fn: Function to run over the data
        max_workers: Maximum number of workers to use
        scheduled_tasks: Number of tasks to schedule
        **threadpool_kwargs: Threadpool keyword arguments

    Returns:
        Iterator of the results of the function over the data
    """
    with futures.ThreadPoolExecutor(max_workers=max_workers, **threadpool_kwargs) as executor:
        futures_deque: deque = deque()
        has_next = True
        itr = iter(data_iterator)
        for _ in range(scheduled_tasks):
            try:
                futures_deque.append(executor.submit(fn, next(itr)))
            except StopIteration:
                has_next = False
                break

        while len(futures_deque) > 0:
            if has_next:
                try:
                    futures_deque.append(executor.submit(fn, next(itr)))
                except StopIteration:
                    has_next = False
            yield futures_deque.popleft().result()



""" Model for splitting data """

from typing import List

from pydantic import BaseModel, validator


class TrainValidationTestSpecific(BaseModel):
    """Class on how to specifically split the data into train, validation and test."""

    train: List[str]
    validation: List[str]
    test: List[str]

    @validator("train")
    def train_validation_test(cls, v, values):
        """Make sure there is no overlap for the train data"""
        for vv in ["test", "validation"]:
            if vv in values.keys():
                overlap = [period for period in v if period in values[vv]]
                if len(overlap):
                    raise ValueError(f"There is a period in both validation and {vv} sets")

        return v

    @validator("validation")
    def validation_overlap(cls, v, values):
        """Make sure there is no overlap for the validation data"""
        for vv in ["test", "train"]:
            if vv in values.keys():
                overlap = [period for period in v if period in values[vv]]
                if len(overlap):
                    raise ValueError(f"There is a period in both validation and {vv} sets")

        return v

    @validator("test")
    def test_overlap(cls, v, values):
        """Make sure there is no overlap for the test data"""
        for vv in ["validation", "train"]:
            if vv in values.keys():
                overlap = [period for period in v if period in values[vv]]
                if len(overlap):
                    raise ValueError(f"There is a period in both test and {vv} sets")

        return v


default_train_test_validation_specific = TrainValidationTestSpecific(
    train=["2015-01-01", "2016-01-01", "2017-01-01", "2018-01-01", "2019-01-01"],
    validation=["2020-01-01"],
    test=["2021-01-01"],
)



"""Select time periods"""

import logging
from typing import Union

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("filter_time_periods")
class FilterTimePeriodsIterDataPipe(IterDataPipe):
    """Select time periods"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        time_periods: IterDataPipe,
        dim_name: str = "time_utc",
    ):
        """
        Select time periods

        Args:
            source_datapipe: Datapipe of Xarray objects
            time_periods: Time periods to select
            dim_name: Dimension name for time
        """
        self.source_datapipe = source_datapipe
        self.time_periods = time_periods
        self.dim_name = dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, time_periods in self.source_datapipe.zip_ocf(self.time_periods):
            new_xr_data = []
            logger.debug(f"Selecting Time periods ({len(time_periods)})")
            for _, row in time_periods.iterrows():
                start_dt = row["start_dt"]
                end_dt = row["end_dt"]
                new_xr_data.append(xr_data.sel({self.dim_name: slice(start_dt, end_dt)}))

            xr_data_concat = xr.concat(new_xr_data, dim=self.dim_name)
            yield xr_data_concat



""" Save functions for the configuration model"""

import json
import logging
from typing import Optional, Union

import fsspec
import yaml
from pathy import Pathy

from ocf_datapipes.config.model import Configuration

logger = logging.getLogger(__name__)


def save_yaml_configuration(
    configuration: Configuration, filename: Optional[Union[str, Pathy]] = None
):
    """
    Save a local yaml file which has the a configuration in it.

    If `filename` is None then saves to configuration.output_data.filepath / configuration.yaml.

    Will save to GCP, AWS, or local, depending on the protocol suffix of filepath.
    """
    # make a dictionary from the configuration,
    # Note that we make the object json'able first, so that it can be saved to a yaml file
    d = json.loads(configuration.json())
    if filename is None:
        filename = Pathy(configuration.output_data.filepath) / "configuration.yaml"
    logger.info(f"Saving configuration to {filename}")

    # save to a yaml file
    with fsspec.open(filename, "w") as yaml_file:
        yaml.safe_dump(d, yaml_file, default_flow_style=False)



"""Encodes the Fourier features for space and time"""

import logging
import warnings
from typing import Union

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch, NWPBatchKey, NWPNumpyBatch

logger = logging.getLogger(__name__)


@functional_datapipe("add_fourier_space_time")
class AddFourierSpaceTimeIterDataPipe(IterDataPipe):
    """Encodes the Fourier features for space and time"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        n_fourier_features_per_dim: int = 8,
    ):
        """
        Encodes the space and time fourier features

        Args:
            source_datapipe: Datapipe of NumpyBatch
            n_fourier_features_per_dim: Number of Fourier features per dimension
        """
        self.source_datapipe = source_datapipe
        self.n_fourier_features_per_dim = n_fourier_features_per_dim

    def __iter__(self):
        for np_batch in self.source_datapipe:
            yield add_spatial_and_temporal_fourier_features(
                np_batch=np_batch,
                n_fourier_features_per_dim=self.n_fourier_features_per_dim,
            )


def add_spatial_and_temporal_fourier_features(
    np_batch: NumpyBatch,
    n_fourier_features_per_dim: int = 8,
) -> NumpyBatch:
    """Add fourier features for x_osgb, y_osgb and time_utc."""
    # NWP is nested so needs to be done separately
    nwp_in_batch = BatchKey.nwp in np_batch
    if nwp_in_batch:
        # Pop and process NWPBatch. Add back in later
        nwp_batch = np_batch.pop(BatchKey.nwp)
        for nwp_source in nwp_batch.keys():
            _add_spatial_and_temporal_fourier_features(
                nwp_batch[nwp_source], n_fourier_features_per_dim
            )

    # Process other coords
    _add_spatial_and_temporal_fourier_features(np_batch, n_fourier_features_per_dim)

    # Add back in NWP maybe
    if nwp_in_batch:
        np_batch[BatchKey.nwp] = nwp_batch

    return np_batch


def _add_spatial_and_temporal_fourier_features(
    batch: Union[NumpyBatch, NWPNumpyBatch],
    n_fourier_features_per_dim: int,
) -> Union[NumpyBatch, NWPNumpyBatch]:
    """Adds fourier encodings in place to batch dict"""
    for key in list(batch.keys()):
        if key.name.endswith(("x_osgb", "y_osgb", "time_utc")):
            if isinstance(key, BatchKey):
                fourier_key = BatchKey[f"{key.name}_fourier"]
            elif isinstance(key, NWPBatchKey):
                fourier_key = NWPBatchKey[f"{key.name}_fourier"]
            else:
                raise ValueError(f"Unregognized key: {key}")

            normalized_coords = normalize_coords(batch[key])

            batch[fourier_key] = compute_fourier_features(
                normalized_coords, n_fourier_features=n_fourier_features_per_dim
            )
    return


def compute_fourier_features(
    array: np.ndarray,
    n_fourier_features: int = 8,
    min_freq: float = 2,
    max_freq: float = 8,
) -> np.ndarray:
    """Compute Fourier features for a single dimension, across all examples in a batch.

    Adapted from https://pytorch.org/tutorials/beginner/transformer_tutorial.html

    Args:
        array: np.ndarray with values roughly in the range [0, 1].
            The values don't have to be *exactly* in the range [0, 1] because sine and cosine
            handle values below 0 and above 2*pi.
            For the time dimension, the shape will be (batch_size, n_timesteps).
            For spatial dimensions, the shape might be (batch_size, length) or
            (batch_size, height, width).
            Although this function can cope with any shape `array`, with any number of dimensions.
        n_fourier_features: Total number of requested Fourier features. Must be an even number
            because half a sine and half are cosine.
        min_freq: If min_freq=2 and array is in the range [0, 1] then the lowest freq "wave" will
            go from -1 to +1 across the dimension.
        max_freq: Maximum frequency for the fourier features

    Returns:
        fourier_features: An np.ndarray of the same dtype as `array`,
            with shape `array.shape + (n_fourier_features,)`. Fourier features with even indexes
            are sine. Odd indexes are cosine.
    """
    assert n_fourier_features % 2 == 0
    assert min_freq > 0
    assert max_freq > min_freq

    div_term = np.linspace(
        start=min_freq,
        stop=max_freq,
        num=n_fourier_features // 2,
        dtype=array.dtype,
    )
    fourier_features = np.full(
        shape=array.shape + (n_fourier_features,),
        fill_value=np.nan,
        dtype=array.dtype,
    )

    radians = array * np.pi / 2
    radians = np.expand_dims(radians, axis=-1)
    radians_x_div_term = radians * div_term
    fourier_features[..., 1::2] = np.cos(radians_x_div_term)
    fourier_features[..., 0::2] = np.sin(radians_x_div_term)

    # Sanity check:
    if np.isfinite(array).all():
        assert np.isfinite(fourier_features).all()
    return fourier_features


def normalize_coords(
    coords: np.ndarray,
) -> np.ndarray:
    """Rescale the coords for a single dimension, across all modalities.

    Args:
        coords: Array of coords
    """
    batch_size = coords.shape[0]

    # min and max taken over these dims
    reduce_dims = tuple(range(1, len(coords.shape)))

    with warnings.catch_warnings():
        # We expect to encounter all-NaN slices (when a whole PV example is missing,
        # for example.)
        warnings.filterwarnings("ignore", "All-NaN slice encountered")

        min_per_example = np.nanmin(
            coords.reshape((batch_size, -1)), axis=reduce_dims, keepdims=True
        )

        max_per_example = np.nanmax(
            coords.reshape((batch_size, -1)), axis=reduce_dims, keepdims=True
        )

    assert np.isfinite(min_per_example).all()
    assert np.isfinite(max_per_example).all()

    normalized_coords = (coords - min_per_example) / (max_per_example - min_per_example)
    return normalized_coords



from ocf_datapipes.transform.xarray import ConvertPressureLevelsToSeparateVariables
import xarray as xr


def test_convert_pressure_levels_to_separate_variables_subselect_levels(
    icon_eu_datapipe,
):
    icon_eu_datapipe = ConvertPressureLevelsToSeparateVariables(
        icon_eu_datapipe, pressure_level_to_use=[50.0]
    )
    data = next(iter(icon_eu_datapipe))
    assert isinstance(data, xr.DataArray)
    for v in ("level", "latitude", "longitude", "step", "init_time_utc"):
        assert v in data.dims
    assert data.shape == (2, 10, 657, 1377, 2)


def test_convert_pressure_levels_to_separate_variables(icon_eu_datapipe):
    icon_eu_datapipe = ConvertPressureLevelsToSeparateVariables(icon_eu_datapipe)
    data = next(iter(icon_eu_datapipe))
    assert isinstance(data, xr.DataArray)
    for v in ("level", "latitude", "longitude", "step", "init_time_utc"):
        assert v in data.dims
    assert data.shape == (2, 10, 657, 1377, 3)


def test_convert_pressure_levels_to_separate_variables_icon_global(icon_global_datapipe):
    icon_global_datapipe = ConvertPressureLevelsToSeparateVariables(icon_global_datapipe)
    data = next(iter(icon_global_datapipe))
    assert isinstance(data, xr.DataArray)
    for v in ("level", "values", "step", "init_time_utc"):
        assert v in data.dims
    assert data.shape == (2, 3, 2949120, 3)



"""Experimental Datapipes

Collection of experimental datapipes, explicitly not ready for production,
but can be upgraded later, these might not be tested, and are to make quick experiments easier
"""



"""Create the training/validation datapipe for training the PVNet Model"""

import logging
from datetime import datetime
from typing import Optional

import xarray as xr
from torch.utils.data.datapipes.datapipe import IterDataPipe

from ocf_datapipes.batch import MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.training.common import (
    _get_datapipes_dict,
    check_nans_in_satellite_data,
    concat_xr_time_utc,
    construct_loctime_pipelines,
    fill_nans_in_arrays,
    fill_nans_in_pv,
    normalize_gsp,
    normalize_pv,
    slice_datapipes_by_time,
)
from ocf_datapipes.utils.consts import (
    NWP_MEANS,
    NWP_STDS,
    RSS_MEAN,
    RSS_STD,
)

xr.set_options(keep_attrs=True)
logger = logging.getLogger("pvnet_datapipe")


def slice_datapipes_by_space(
    datapipes_dict: dict,
    location_pipe: IterDataPipe,
    configuration: Configuration,
) -> None:
    """Modifies a dictionary of datapipes in-place to yield slices around a given location

    Args:
        datapipes_dict: Dictionary of used datapipes and t0 ones
        location_pipe: Datapipe which yields location for sample
        configuration: Configuration object.
    """

    # Unpack for convenience
    conf_sat = configuration.input_data.satellite
    conf_nwp = configuration.input_data.nwp

    if "nwp" in datapipes_dict:
        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
            datapipes_dict["nwp"][nwp_key] = nwp_datapipe.select_spatial_slice_pixels(
                location_pipe_copy,
                roi_height_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_height,
                roi_width_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_width,
            )

    if "sat" in datapipes_dict:
        location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
        datapipes_dict["sat"] = datapipes_dict["sat"].select_spatial_slice_pixels(
            location_pipe_copy,
            roi_height_pixels=conf_sat.satellite_image_size_pixels_height,
            roi_width_pixels=conf_sat.satellite_image_size_pixels_width,
        )

    # GSP always assumed to be in data
    datapipes_dict["gsp"] = datapipes_dict["gsp"].select_spatial_slice_meters(
        location_datapipe=location_pipe,
        roi_height_meters=1,
        roi_width_meters=1,
        dim_name="gsp_id",
    )

    return


def process_and_combine_datapipes(datapipes_dict, configuration, check_satellite_no_nans=False):
    """Normalize and convert data to numpy arrays"""

    # Unpack for convenience
    conf_nwp = configuration.input_data.nwp

    numpy_modalities = []

    # Normalise the inputs and convert to numpy format
    if "nwp" in datapipes_dict:
        nwp_numpy_modalities = dict()

        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            nwp_datapipe = nwp_datapipe.normalize(
                mean=NWP_MEANS[conf_nwp[nwp_key].nwp_provider],
                std=NWP_STDS[conf_nwp[nwp_key].nwp_provider],
            )
            nwp_numpy_modalities[nwp_key] = nwp_datapipe.convert_nwp_to_numpy_batch()

        # Combine the NWPs into NumpyBatch
        nwp_numpy_modalities = MergeNWPNumpyModalities(nwp_numpy_modalities)
        numpy_modalities.append(nwp_numpy_modalities)

    if "sat" in datapipes_dict:
        sat_datapipe = datapipes_dict["sat"]
        sat_datapipe = sat_datapipe.normalize(mean=RSS_MEAN, std=RSS_STD)
        # Check for large amount of zeros
        sat_datapipe = sat_datapipe.check_value_equal_to_fraction(value=0.0, fraction=0.9)
        numpy_modalities.append(sat_datapipe.convert_satellite_to_numpy_batch())

    if "pv" in datapipes_dict:
        # Recombine PV arrays - see function doc for further explanation
        pv_datapipe = (
            datapipes_dict["pv"].zip_ocf(datapipes_dict["pv_future"]).map(concat_xr_time_utc)
        )
        pv_datapipe = pv_datapipe.normalize(normalize_fn=normalize_pv)
        pv_datapipe = pv_datapipe.map(fill_nans_in_pv)

        numpy_modalities.append(pv_datapipe.convert_pv_to_numpy_batch())

    # GSP always assumed to be in data
    gsp_future_datapipe = datapipes_dict["gsp_future"]
    gsp_datapipe = datapipes_dict["gsp"]

    # Recombine GSP arrays - see function doc for further explanation
    gsp_datapipe = gsp_datapipe.zip_ocf(gsp_future_datapipe).map(concat_xr_time_utc)
    gsp_datapipe = gsp_datapipe.normalize(normalize_fn=normalize_gsp)

    numpy_modalities.append(gsp_datapipe.convert_gsp_to_numpy_batch())

    logger.debug("Combine all the data sources")
    combined_datapipe = MergeNumpyModalities(numpy_modalities).add_sun_position(modality_name="gsp")

    if check_satellite_no_nans:
        # in production we don't want any nans in the satellite data
        combined_datapipe = combined_datapipe.map(check_nans_in_satellite_data)

    combined_datapipe = combined_datapipe.map(fill_nans_in_arrays)

    return combined_datapipe


def construct_sliced_data_pipeline(
    config_filename: str,
    location_pipe: IterDataPipe,
    t0_datapipe: IterDataPipe,
    production: bool = False,
    check_satellite_no_nans: bool = False,
) -> IterDataPipe:
    """Constructs data pipeline for the input data config file.

    This yields samples from the location and time datapipes.

    Args:
        config_filename: Path to config file.
        location_pipe: Datapipe yielding locations.
        t0_datapipe: Datapipe yielding times.
        production: Whether constucting pipeline for production inference.
        check_satellite_no_nans: Whether to check that satellite data has no nans.
    """

    datapipes_dict = _get_datapipes_dict(
        config_filename,
        production=production,
    )

    configuration = datapipes_dict.pop("config")

    # Slice all of the datasets by space - this is an in-place operation
    slice_datapipes_by_space(datapipes_dict, location_pipe, configuration)

    # Slice all of the datasets by time - this is an in-place operation
    slice_datapipes_by_time(datapipes_dict, t0_datapipe, configuration, production)

    # Normalise, and combine the data sources into NumpyBatches
    combined_datapipe = process_and_combine_datapipes(
        datapipes_dict, configuration, check_satellite_no_nans
    )

    return combined_datapipe


def pvnet_datapipe(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> IterDataPipe:
    """
    Construct pvnet pipeline for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time at which a sample can be selected.
        end_time: Maximum time at which a sample can be selected.
    """
    logger.info("Constructing pvnet pipeline")

    # Open datasets from the config and filter to useable location-time pairs
    location_pipe, t0_datapipe = construct_loctime_pipelines(
        config_filename,
        start_time,
        end_time,
    )

    # Shard after we have the loc-times. These are already shuffled so no need to shuffle again
    location_pipe = location_pipe.sharding_filter()
    t0_datapipe = t0_datapipe.sharding_filter()

    # In this function we re-open the datasets to make a clean separation before/after sharding
    # This function
    datapipe = construct_sliced_data_pipeline(
        config_filename,
        location_pipe,
        t0_datapipe,
    )

    return datapipe



from ocf_datapipes.transform.xarray import Upsample


def test_nwp_upsample(nwp_datapipe):
    nwp_datapipe = Upsample(
        nwp_datapipe, y_upsample=2, x_upsample=2, x_dim_name="x_osgb", y_dim_name="y_osgb"
    )
    data = next(iter(nwp_datapipe))

    # Upsample by 2 from 704x548
    assert data.shape[-1] == 548 * 2 - 1
    assert data.shape[-2] == 704 * 2 - 1


def test_sat_downsample(sat_datapipe):
    sat_datapipe = Upsample(
        sat_datapipe,
        y_upsample=2,
        x_upsample=2,
        y_dim_name="y_geostationary",
        x_dim_name="x_geostationary",
    )
    data = next(iter(sat_datapipe))
    assert data.shape[-1] == 615 * 2
    assert data.shape[-2] == 298 * 2 - 1


def test_nwp_upsample_keep_same_shape(nwp_datapipe):
    nwp_datapipe_new, nwp_datapipe_old = nwp_datapipe.fork(2)

    nwp_datapipe_new = Upsample(
        nwp_datapipe_new,
        y_upsample=2,
        x_upsample=2,
        x_dim_name="x_osgb",
        y_dim_name="y_osgb",
        keep_same_shape=True,
    )
    data_new = next(iter(nwp_datapipe_new))
    data_old = next(iter(nwp_datapipe_old))

    assert data_new.shape[-1] == 548
    assert data_new.shape[-2] == 704

    # check first values are different
    assert data_new.x_osgb.values[0] != data_old.x_osgb.values[0]
    assert data_new.y_osgb.values[0] != data_old.y_osgb.values[0]

    # check middle valyes are the same
    assert data_new.x_osgb.values[274] == data_old.x_osgb.values[274]
    assert data_new.y_osgb.values[351] == data_old.y_osgb.values[352]



"""Convert PV to Numpy"""

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import NumpyBatch


@functional_datapipe("convert_pv_to_numpy")
class ConvertPVToNumpyIterDataPipe(IterDataPipe):
    """Convert PV Xarray to NumpyBatch"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        return_pv_id: bool = False,
        return_ml_id: bool = False,
    ):
        """
        Convert PV Xarray objects to NumpyBatch objects

        Args:
            source_datapipe: Datapipe emitting PV Xarray objects
            return_pv_id: Whether to return PV ID
            return_ml_id: Whether to return ML ID
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.return_pv_id = return_pv_id
        self.return_ml_id = return_ml_id

    def __iter__(self) -> NumpyBatch:
        """Iterate and convert PV Xarray to NumpyBatch"""
        for xr_data in self.source_datapipe:
            pv_yield_history = xr_data.values
            returned_values = [pv_yield_history]
            if self.return_ml_id:
                ml_ids = xr_data["ml_id"].values.astype(np.float32)
                returned_values.append(ml_ids)
            if self.return_pv_id:
                pv_id = xr_data["pv_system_id"].values.astype(np.float32)
                returned_values.append(pv_id)
            yield returned_values



"""OCF Datapipes"""

import ocf_datapipes.batch
import ocf_datapipes.convert
import ocf_datapipes.experimental
import ocf_datapipes.load
import ocf_datapipes.select
import ocf_datapipes.transform
import ocf_datapipes.utils
import ocf_datapipes.validation



"""Reproject Topographic data to OSGB"""

from pathlib import Path
from typing import Union

import numpy as np
import pyproj
import pyresample
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

try:
    import cartopy.crs as ccrs

    _has_cartopy = True
except ImportError:
    _has_cartopy = False


@functional_datapipe("reproject_topography")
class ReprojectTopographyIterDataPipe(IterDataPipe):
    """Reproject Topographic data to OSGB"""

    def __init__(self, topo_datapipe: Union[Path, str]):
        """
        Reproject topo data to OSGB

        Args:
            topo_datapipe: Datapipe emitting topographic data
        """
        self.topo_datapipe = topo_datapipe

    def __iter__(self) -> xr.DataArray:
        """Reproject topographic data"""
        for topo in self.topo_datapipe:
            # Select Western Europe:
            topo = topo.sel(x_osgb=slice(-300_000, 1_500_000), y_osgb=slice(1_300_000, -800_000))

            topo = reproject_topo_data_from_osgb_to_geostationary(topo)
            topo = topo.fillna(0)

            while True:
                yield topo


def reproject_topo_data_from_osgb_to_geostationary(topo: xr.DataArray) -> xr.DataArray:
    """
    Reproject topographic data from osgb to geostationary

    Args:
        topo: Topographic Xarray DataArray

    Returns:
        Reprojected Topographic Xarray DataArray
    """
    topo_osgb_area_def = _get_topo_osgb_area_def(topo)
    topo_geostationary_area_def = _get_topo_geostationary_area_def(topo)
    topo_image = pyresample.image.ImageContainerQuick(topo.values, topo_osgb_area_def)
    topo_image_resampled = topo_image.resample(topo_geostationary_area_def)
    topo_dataarray = _get_data_array_of_resampled_topo_image(topo_image_resampled)
    # topo_dataarray.attrs["area"] = str(topo_geostationary_area_def)
    return topo_dataarray


def _get_topo_osgb_area_def(topo: xr.DataArray) -> pyresample.geometry.AreaDefinition:
    """
    Get the area definition for resampling for OSGB

    Args:
        topo: Topographic Xarray DataArray

    Returns:
        AreaDefinition for OSGB
    """
    # Get AreaDefinition of the OSGB topographical data:
    if not _has_cartopy:
        raise Exception("Please install `cartopy` before using ReprojectTopography")
    osgb = ccrs.OSGB(approx=False)
    return pyresample.create_area_def(
        area_id="OSGB",
        projection=osgb.proj4_params,
        shape=topo.shape,  # y, x
        area_extent=(
            topo.x_osgb[0].item(),  # lower_left_x
            topo.y_osgb[-1].item(),  # lower_left_y
            topo.x_osgb[-1].item(),  # upper_right_x
            topo.y_osgb[0].item(),  # upper_right_y
        ),
    )


def _get_topo_geostationary_area_def(topo: xr.DataArray) -> pyresample.geometry.AreaDefinition:
    """
    Get the geostationary area definition for Topographic data

    Args:
        topo: Topographic Xarray DataArray

    Returns:
        AreaDefinition for geostationary area of RSS imagery
    """
    # Get the geostationary boundaries of the topo data:
    OSGB_EPSG_CODE = 27700
    GEOSTATIONARY_PROJ = {
        "proj": "geos",
        "lon_0": 9.5,
        "h": 35785831,
        "x_0": 0,
        "y_0": 0,
        "a": 6378169,
        "rf": 295.488065897014,
        "no_defs": None,
        "type": "crs",
    }
    osgb_to_geostationary = pyproj.Transformer.from_crs(
        crs_from=OSGB_EPSG_CODE, crs_to=GEOSTATIONARY_PROJ
    ).transform
    lower_left_geos = osgb_to_geostationary(xx=topo.x_osgb[0], yy=topo.y_osgb[-1])
    upper_right_geos = osgb_to_geostationary(xx=topo.x_osgb[-1], yy=topo.y_osgb[0])
    shape = (topo.shape[0] * 2, topo.shape[1] * 2)  # Oversample to make sure we don't loose info.
    return pyresample.create_area_def(
        area_id="msg_seviri_rss_1km",
        projection=GEOSTATIONARY_PROJ,
        shape=shape,
        # lower_left_x, lower_left_y, upper_right_x, upper_right_y:
        area_extent=lower_left_geos + upper_right_geos,
    )


def _get_data_array_of_resampled_topo_image(
    topo_image_resampled: pyresample.image.ImageContainer,
) -> xr.DataArray:
    """
    Put resampled data into a DataArray

    Args:
        topo_image_resampled: Resampled data

    Returns:
        Xarray DataArray containing the resampled data
    """
    (
        lower_left_x,
        lower_left_y,
        upper_right_x,
        upper_right_y,
    ) = topo_image_resampled.geo_def.area_extent
    return xr.DataArray(
        topo_image_resampled.image_data,
        coords=(
            (
                "y",
                np.linspace(
                    start=upper_right_y, stop=lower_left_y, num=topo_image_resampled.shape[0]
                ),
            ),
            (
                "x",
                np.linspace(
                    start=lower_left_x, stop=upper_right_x, num=topo_image_resampled.shape[1]
                ),
            ),
        ),
    )



"""Convert GSP to Numpy Array"""

import logging

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("convert_gsp_to_numpy")
class ConvertGSPToNumpyIterDataPipe(IterDataPipe):
    """Convert GSP Xarray to Numpy Array"""

    def __init__(self, source_datapipe: IterDataPipe, return_id: bool = False):
        """
        Convert GSP Xarray to Numpy Array object

        Args:
            source_datapipe: Datapipe emitting GSP Xarray object
            return_id: Return GSP ID as well
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.return_id = return_id

    def __iter__(self) -> np.ndarray:
        """Convert from Xarray to Numpy array"""
        logger.debug("Converting GSP to numpy")
        for xr_data in self.source_datapipe:
            returned_values = [xr_data.values]
            if self.return_id:
                pv_system_ids = xr_data["gsp_id"].values
                returned_values.append(pv_system_ids)
                yield returned_values
            else:
                yield returned_values[0]



"""GSP Loader"""

import datetime
import logging
from pathlib import Path
from typing import Union

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("open_gsp_national")
class OpenGSPNationalIterDataPipe(IterDataPipe):
    """Get and open the GSP data"""

    def __init__(
        self,
        gsp_pv_power_zarr_path: Union[str, Path],
        sample_period_duration: datetime.timedelta = datetime.timedelta(minutes=30),
    ):
        """
        Get and open the GSP data

        Args:
            gsp_pv_power_zarr_path: Path to the Zarr for GSP PV Power
            sample_period_duration: Sample period of the GSP data
        """
        self.gsp_pv_power_zarr_path = gsp_pv_power_zarr_path
        self.sample_period_duration = sample_period_duration

    def __iter__(self) -> xr.DataArray:
        """Get and return GSP data"""

        logger.debug("Getting GSP data")

        # Load GSP generation xr.Dataset:
        gsp_pv_power_mw_ds = xr.load_dataset(self.gsp_pv_power_zarr_path, engine="zarr")

        # only select national data
        logger.debug("Selecting National data only")
        gsp_pv_power_mw_ds = gsp_pv_power_mw_ds.sel(gsp_id=0)

        # rename some variables
        data_array = gsp_pv_power_mw_ds.rename(
            {
                "datetime_gmt": "time_utc",
                "generation_mw": "gsp_pv_power_mw",
                "installedcapacity_mwp": "nominal_capacity_mwp",
                "capacity_mwp": "effective_capacity_mwp",
            }
        )

        while True:
            yield data_array



"""Select spatial slices"""

import logging
from typing import Optional, Union

import numpy as np
import xarray as xr
from scipy.spatial import KDTree
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.utils import Location
from ocf_datapipes.utils.geospatial import (
    lon_lat_to_geostationary_area_coords,
    lon_lat_to_osgb,
    move_lon_lat_by_meters,
    osgb_to_geostationary_area_coords,
    osgb_to_lon_lat,
    spatial_coord_type,
)
from ocf_datapipes.utils.utils import searchsorted

logger = logging.getLogger(__name__)


# -------------------------------- utility functions --------------------------------


def convert_coords_to_match_xarray(x, y, from_coords, xr_data):
    """Convert x and y coords to cooridnate system matching xarray data

    Args:
        x: Float or array-like
        y: Float or array-like
        from_coords: String describing coordinate system of x and y
        xr_data: xarray data object to which coordinates should be matched
    """

    xr_coords, xr_x_dim, xr_y_dim = spatial_coord_type(xr_data)

    assert from_coords in ["osgb", "lon_lat"]
    assert xr_coords in ["geostationary", "osgb", "lon_lat"]

    if xr_coords == "geostationary":
        if from_coords == "osgb":
            x, y = osgb_to_geostationary_area_coords(x, y, xr_data)

        elif from_coords == "lon_lat":
            x, y = lon_lat_to_geostationary_area_coords(x, y, xr_data)

    elif xr_coords == "lon_lat":
        if from_coords == "osgb":
            x, y = osgb_to_lon_lat(x, y)

        # else the from_coords=="lon_lat" and we don't need to convert

    elif xr_coords == "osgb":
        if from_coords == "lon_lat":
            x, y = lon_lat_to_osgb(x, y)

        # else the from_coords=="osgb" and we don't need to convert

    return x, y


def _get_idx_of_pixel_closest_to_poi(
    xr_data: xr.DataArray,
    location: Location,
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        location: Center
    Returns:
        The Location for the center pixel
    """
    xr_coords, xr_x_dim, xr_y_dim = spatial_coord_type(xr_data)

    if xr_coords not in ["osgb", "lon_lat"]:
        raise NotImplementedError(f"Only 'osgb' and 'lon_lat' are supported - not '{xr_coords}'")

    # Convert location coords to match xarray data
    x, y = convert_coords_to_match_xarray(
        location.x,
        location.y,
        from_coords=location.coordinate_system,
        xr_data=xr_data,
    )

    # Check that the requested point lies within the data
    assert xr_data[xr_x_dim].min() < x < xr_data[xr_x_dim].max()
    assert xr_data[xr_y_dim].min() < y < xr_data[xr_y_dim].max()

    x_index = xr_data.get_index(xr_x_dim)
    y_index = xr_data.get_index(xr_y_dim)

    closest_x = x_index.get_indexer([x], method="nearest")[0]
    closest_y = y_index.get_indexer([y], method="nearest")[0]

    return Location(x=closest_x, y=closest_y, coordinate_system="idx")


def _get_idx_of_pixel_closest_to_poi_geostationary(
    xr_data: xr.DataArray,
    center_osgb: Location,
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        center_osgb: Center in OSGB coordinates

    Returns:
        Location for the center pixel in geostationary coordinates
    """

    xr_coords, xr_x_dim, xr_y_dim = spatial_coord_type(xr_data)

    x, y = osgb_to_geostationary_area_coords(x=center_osgb.x, y=center_osgb.y, xr_data=xr_data)
    center_geostationary = Location(x=x, y=y, coordinate_system="geostationary")

    # Check that the requested point lies within the data
    assert xr_data[xr_x_dim].min() < x < xr_data[xr_x_dim].max()
    assert xr_data[xr_y_dim].min() < y < xr_data[xr_y_dim].max()

    # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
    x_index_at_center = searchsorted(
        xr_data[xr_x_dim].values, center_geostationary.x, assume_ascending=True
    )

    # y_geostationary is in descending order:
    y_index_at_center = searchsorted(
        xr_data[xr_y_dim].values, center_geostationary.y, assume_ascending=False
    )

    return Location(x=x_index_at_center, y=y_index_at_center, coordinate_system="idx")


def _get_points_from_unstructured_grids(
    xr_data: xr.DataArray,
    location: Location,
    location_idx_name: str = "values",
    num_points: int = 1,
):
    """
    Get the closest points from an unstructured grid (i.e. Icosahedral grid)

    This is primarily used for the Icosahedral grid, which is not a regular grid,
     and so is not an image

    Args:
        xr_data: Xarray dataset
        location: Location of center point
        location_idx_name: Name of the index values dimension
            (i.e. where we index into to get the lat/lon for that point)
        num_points: Number of points to return (should be width * height)

    Returns:
        The closest points from the grid
    """
    xr_coords, xr_x_dim, xr_y_dim = spatial_coord_type(xr_data)
    assert xr_coords == "lon_lat"

    # Check if need to convert from different coordinate system to lat/lon
    if location.coordinate_system == "osgb":
        longitude, latitude = osgb_to_lon_lat(x=location.x, y=location.y)
        location = Location(
            x=longitude,
            y=latitude,
            coordinate_system="lon_lat",
        )
    elif location.coordinate_system == "geostationary":
        raise NotImplementedError(
            "Does not currently support geostationary coordinates when using unstructured grids"
        )

    # Extract lat, lon, and locidx data
    lat = xr_data.longitude.values
    lon = xr_data.latitude.values
    locidx = xr_data[location_idx_name].values

    # Create a KDTree
    tree = KDTree(list(zip(lat, lon)))

    # Query with the [longitude, latitude] of your point
    _, idx = tree.query([location.x, location.y], k=num_points)

    # Retrieve the location_idxs for these grid points
    location_idxs = locidx[idx]

    data = xr_data.sel({location_idx_name: location_idxs})
    return data


# ---------------------------- sub-functions for slicing ----------------------------


def _slice_patial_spatial_pixel_window_from_xarray(
    xr_data,
    left_idx,
    right_idx,
    top_idx,
    bottom_idx,
    left_pad_pixels,
    right_pad_pixels,
    top_pad_pixels,
    bottom_pad_pixels,
    xr_x_dim,
    xr_y_dim,
):
    """Return spatial window of given pixel size when window partially overlaps input data"""

    dx = np.median(np.diff(xr_data[xr_x_dim].values))
    dy = np.median(np.diff(xr_data[xr_y_dim].values))

    if left_pad_pixels > 0:
        assert right_pad_pixels == 0
        x_sel = np.concatenate(
            [
                xr_data[xr_x_dim].values[0] - np.arange(left_pad_pixels, 0, -1) * dx,
                xr_data[xr_x_dim].values[0:right_idx],
            ]
        )
        xr_data = xr_data.isel({xr_x_dim: slice(0, right_idx)}).reindex({xr_x_dim: x_sel})

    elif right_pad_pixels > 0:
        assert left_pad_pixels == 0
        x_sel = np.concatenate(
            [
                xr_data[xr_x_dim].values[left_idx:],
                xr_data[xr_x_dim].values[-1] + np.arange(1, right_pad_pixels + 1) * dx,
            ]
        )
        xr_data = xr_data.isel({xr_x_dim: slice(left_idx, None)}).reindex({xr_x_dim: x_sel})

    else:
        xr_data = xr_data.isel({xr_x_dim: slice(left_idx, right_idx)})

    if top_pad_pixels > 0:
        assert bottom_pad_pixels == 0
        y_sel = np.concatenate(
            [
                xr_data[xr_y_dim].values[0] - np.arange(top_pad_pixels, 0, -1) * dy,
                xr_data[xr_y_dim].values[0:bottom_idx],
            ]
        )
        xr_data = xr_data.isel({xr_y_dim: slice(0, bottom_idx)}).reindex({xr_y_dim: y_sel})

    elif bottom_pad_pixels > 0:
        assert top_pad_pixels == 0
        y_sel = np.concatenate(
            [
                xr_data[xr_y_dim].values[top_idx:],
                xr_data[xr_y_dim].values[-1] + np.arange(1, bottom_pad_pixels + 1) * dy,
            ]
        )
        xr_data = xr_data.isel({xr_y_dim: slice(top_idx, None)}).reindex({xr_x_dim: y_sel})

    else:
        xr_data = xr_data.isel({xr_y_dim: slice(top_idx, bottom_idx)})

    return xr_data


def slice_spatial_pixel_window_from_xarray(
    xr_data, center_idx, width_pixels, height_pixels, xr_x_dim, xr_y_dim, allow_partial_slice
):
    """Select a spatial slice from an xarray object

    Args:
        xr_data: Xarray object
        center_idx: Location object describing the centre of the window
        width_pixels: Window with in pixels
        height_pixels: Window height in pixels
        xr_x_dim: Name of the x-dimension in the xr_data
        xr_y_dim: Name of the y-dimension in the xr_data
        allow_partial_slice: Whether to allow a partially filled window
    """
    half_width = width_pixels // 2
    half_height = height_pixels // 2

    left_idx = int(center_idx.x - half_width)
    right_idx = int(center_idx.x + half_width)
    top_idx = int(center_idx.y - half_height)
    bottom_idx = int(center_idx.y + half_height)

    data_width_pixels = len(xr_data[xr_x_dim])
    data_height_pixels = len(xr_data[xr_y_dim])

    left_pad_required = left_idx < 0
    right_pad_required = right_idx >= data_width_pixels
    top_pad_required = top_idx < 0
    bottom_pad_required = bottom_idx >= data_height_pixels

    pad_required = any(
        [left_pad_required, right_pad_required, top_pad_required, bottom_pad_required]
    )

    if pad_required:
        if allow_partial_slice:
            left_pad_pixels = (-left_idx) if left_pad_required else 0
            right_pad_pixels = (right_idx - (data_width_pixels - 1)) if right_pad_required else 0
            top_pad_pixels = (-top_idx) if top_pad_required else 0
            bottom_pad_pixels = (
                (bottom_idx - (data_height_pixels - 1)) if bottom_pad_required else 0
            )

            xr_data = _slice_patial_spatial_pixel_window_from_xarray(
                xr_data,
                left_idx,
                right_idx,
                top_idx,
                bottom_idx,
                left_pad_pixels,
                right_pad_pixels,
                top_pad_pixels,
                bottom_pad_pixels,
                xr_x_dim,
                xr_y_dim,
            )
        else:
            raise ValueError(
                f"Window for location {center_idx} not available. Missing (left, right, top, "
                f"bottom) pixels  = ({left_pad_required}, {right_pad_required}, "
                f"{top_pad_required}, {bottom_pad_required}). "
                f"You may wish to set `allow_partial_slice=True`"
            )

    else:
        xr_data = xr_data.isel(
            {
                xr_x_dim: slice(left_idx, right_idx),
                xr_y_dim: slice(top_idx, bottom_idx),
            }
        )

    assert len(xr_data[xr_x_dim]) == width_pixels, (
        f"Expected x-dim len {width_pixels} got {len(xr_data[xr_x_dim])} "
        f"for location {center_idx} for slice {left_idx}:{right_idx}"
    )
    assert len(xr_data[xr_y_dim]) == height_pixels, (
        f"Expected y-dim len {height_pixels} got {len(xr_data[xr_y_dim])} "
        f"for location {center_idx} for slice {top_idx}:{bottom_idx}"
    )

    return xr_data


# ---------------------------- main functions for slicing ---------------------------


def select_spatial_slice_pixels(
    xr_data: Union[xr.Dataset, xr.DataArray],
    location: Location,
    roi_width_pixels: int,
    roi_height_pixels: int,
    allow_partial_slice: bool = False,
    location_idx_name: Optional[str] = None,
):
    """
    Select spatial slice based off pixels from location point of interest

    If `allow_partial_slice` is set to True, then slices may be made which intersect the border
    of the input data. The additional x and y cordinates that would be required for this slice
    are extrapolated based on the average spacing of these coordinates in the input data.
    However, currently slices cannot be made where the centre of the window is outside of the
    input data.

    Args:
        xr_data: Xarray DataArray or Dataset to slice from
        location: Location of interest
        roi_height_pixels: ROI height in pixels
        roi_width_pixels: ROI width in pixels
        allow_partial_slice: Whether to allow a partial slice.
        location_idx_name: Name for location index of unstructured grid data,
            None if not relevant
    """

    xr_coords, xr_x_dim, xr_y_dim = spatial_coord_type(xr_data)
    if location_idx_name is not None:
        selected = _get_points_from_unstructured_grids(
            xr_data=xr_data,
            location=location,
            location_idx_name=location_idx_name,
            num_points=roi_width_pixels * roi_height_pixels,
        )
    else:
        if xr_coords == "geostationary":
            center_idx: Location = _get_idx_of_pixel_closest_to_poi_geostationary(
                xr_data=xr_data,
                center_osgb=location,
            )
        else:
            center_idx: Location = _get_idx_of_pixel_closest_to_poi(
                xr_data=xr_data,
                location=location,
            )

        selected = slice_spatial_pixel_window_from_xarray(
            xr_data,
            center_idx,
            roi_width_pixels,
            roi_height_pixels,
            xr_x_dim,
            xr_y_dim,
            allow_partial_slice=allow_partial_slice,
        )

    return selected


def select_spatial_slice_meters(
    xr_data: Union[xr.Dataset, xr.DataArray],
    location: Location,
    roi_width_meters: int,
    roi_height_meters: int,
    dim_name: Optional[str] = None,
):
    """
    Select spatial slice based off pixels from point of interest

    Args:
        xr_data: Xarray DataArray or Dataset to slice from
        location: Location of interest
        roi_height_meters: ROI height in meters
        roi_width_meters: ROI width in meters
        dim_name: Dimension name to select for ID, None for coordinates

    Notes:
        Using spatial slicing based on distance rather than number of pixels will often yield
        slices which can vary by 1 pixel in height and/or width.

        E.g. Suppose the Xarray data has x-coords = [1,2,3,4,5]. We want to slice a spatial
        window with a size which equates to 2.2 along the x-axis. If we choose to slice around
        the point x=3 this will slice out the x-coords [2,3,4]. If we choose to slice around the
        point x=2.5 this will slice out the x-coords [2,3]. Hence the returned slice can have
        size either 2 or 3 in the x-axis depending on the spatial location selected.

        Also, if selecting over a large span of latitudes, this may also causes pixel sizes of
        the yielded outputs to change. For example, if the Xarray data is on a regularly spaced
        longitude-latitude grid, then the structure of the grid means that the longitudes near
        to the poles are spaced closer together (measured in meters) than at the equator. So
        slices near the equator will have less pixels in the x-axis than slices taken near the
        poles.
    """
    # Get the spatial coords of the xarray data
    xr_coords, xr_x_dim, xr_y_dim = spatial_coord_type(xr_data)

    half_width = roi_width_meters // 2
    half_height = roi_height_meters // 2

    # Find the bounding box values for the location in either lat-lon or OSGB coord systems
    if location.coordinate_system == "lon_lat":
        right, top = move_lon_lat_by_meters(
            location.x,
            location.y,
            half_width,
            half_height,
        )
        left, bottom = move_lon_lat_by_meters(
            location.x,
            location.y,
            -half_width,
            -half_height,
        )

    elif location.coordinate_system == "osgb":
        left = location.x - half_width
        right = location.x + half_width
        bottom = location.y - half_height
        top = location.y + half_height

    else:
        raise ValueError(f"Location coord system not recognized: {location.coordinate_system}")

    # Change the bounding coordinates [left, right, bottom, top] to the same
    # coordinate system as the xarray data
    (left, right), (bottom, top) = convert_coords_to_match_xarray(
        x=np.array([left, right], dtype=np.float32),
        y=np.array([bottom, top], dtype=np.float32),
        from_coords=location.coordinate_system,
        xr_data=xr_data,
    )

    # Do it off coordinates, not ID
    if dim_name is None:
        # Select a patch from the xarray data
        x_mask = (left <= xr_data[xr_x_dim]) & (xr_data[xr_x_dim] <= right)
        y_mask = (bottom <= xr_data[xr_y_dim]) & (xr_data[xr_y_dim] <= top)
        selected = xr_data.isel({xr_x_dim: x_mask, xr_y_dim: y_mask})

    else:
        # Select data in the region of interest and ID:
        # This also works for unstructured grids

        id_mask = (
            (left <= xr_data[xr_x_dim])
            & (xr_data[xr_x_dim] <= right)
            & (bottom <= xr_data[xr_y_dim])
            & (xr_data[xr_y_dim] <= top)
        )
        selected = xr_data.isel({dim_name: id_mask})
    return selected


# ------------------------------ datapipes for slicing ------------------------------


@functional_datapipe("select_spatial_slice_pixels")
class SelectSpatialSlicePixelsIterDataPipe(IterDataPipe):
    """Select spatial slice based off pixels from point of interest"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        location_datapipe: IterDataPipe,
        roi_width_pixels: int,
        roi_height_pixels: int,
        allow_partial_slice: bool = False,
        location_idx_name: Optional[str] = None,
    ):
        """
        Select spatial slice based off pixels from point of interest

        If `allow_partial_slice` is set to True, then slices may be made which intersect the border
        of the input data. The additional x and y cordinates that would be required for this slice
        are extrapolated based on the average spacing of these coordinates in the input data.
        However, currently slices cannot be made where the centre of the window is outside of the
        input data.

        Args:
            source_datapipe: Datapipe of Xarray data
            location_datapipe: Location datapipe
            roi_height_pixels: ROI height in pixels
            roi_width_pixels: ROI width in pixels
            allow_partial_slice: Whether to allow a partial slice.
            location_idx_name: Name for location index of unstructured grid data,
                None if not relevant
        """
        self.source_datapipe = source_datapipe
        self.location_datapipe = location_datapipe
        self.roi_height_pixels = roi_height_pixels
        self.roi_width_pixels = roi_width_pixels
        self.allow_partial_slice = allow_partial_slice
        self.location_idx_name = location_idx_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, location in self.source_datapipe.zip_ocf(self.location_datapipe):
            logger.debug("Selecting spatial slice with pixels")

            selected = select_spatial_slice_pixels(
                xr_data=xr_data,
                location=location,
                roi_width_pixels=self.roi_width_pixels,
                roi_height_pixels=self.roi_height_pixels,
                allow_partial_slice=self.allow_partial_slice,
                location_idx_name=self.location_idx_name,
            )

            yield selected


@functional_datapipe("select_spatial_slice_meters")
class SelectSpatialSliceMetersIterDataPipe(IterDataPipe):
    """Select spatial slice based off meters from point of interest"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        location_datapipe: IterDataPipe,
        roi_width_meters: int,
        roi_height_meters: int,
        dim_name: Optional[str] = None,
    ):
        """
        Select spatial slice based off pixels from point of interest

        Args:
            source_datapipe: Datapipe of Xarray data
            location_datapipe: Location datapipe
            roi_width_meters: ROI width in meters
            roi_height_meters: ROI height in meters
            dim_name: Dimension name to select for ID, None for coordinates

        Notes:
            Using spatial slicing based on distance rather than number of pixels will often yield
            slices which can vary by 1 pixel in height and/or width.

            E.g. Suppose the Xarray data has x-coords = [1,2,3,4,5]. We want to slice a spatial
            window with a size which equates to 2.2 along the x-axis. If we choose to slice around
            the point x=3 this will slice out the x-coords [2,3,4]. If we choose to slice around the
            point x=2.5 this will slice out the x-coords [2,3]. Hence the returned slice can have
            size either 2 or 3 in the x-axis depending on the spatial location selected.

            Also, if selecting over a large span of latitudes, this may also causes pixel sizes of
            the yielded outputs to change. For example, if the Xarray data is on a regularly spaced
            longitude-latitude grid, then the structure of the grid means that the longitudes near
            to the poles are spaced closer together (measured in meters) than at the equator. So
            slices near the equator will have less pixels in the x-axis than slices taken near the
            poles.
        """
        self.source_datapipe = source_datapipe
        self.location_datapipe = location_datapipe
        self.roi_height_meters = roi_height_meters
        self.roi_width_meters = roi_width_meters
        self.dim_name = dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data, location in self.source_datapipe.zip_ocf(self.location_datapipe):
            # Compute the index for left and right:
            logger.debug("Getting Spatial Slice Meters")

            selected = select_spatial_slice_meters(
                xr_data=xr_data,
                location=location,
                roi_width_meters=self.roi_width_meters,
                roi_height_meters=self.roi_height_meters,
                dim_name=self.dim_name,
            )

            yield selected



""" Data pipes and utils for getting PV data from database"""

import logging
import os
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from nowcasting_datamodel.connection import DatabaseConnection
from nowcasting_datamodel.models.base import Base_Forecast
from sqlalchemy import text
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.load.pv.utils import put_pv_data_into_an_xr_dataarray

logger = logging.getLogger(__name__)


@functional_datapipe("open_pv_from_pvsites_db")
class OpenPVFromPVSitesDBIterDataPipe(IterDataPipe):
    """Data pipes and utils for getting PV data from pvsites database"""

    def __init__(
        self,
        history_minutes: int = 30,
    ):
        """
        Datapipe to get PV from pvsites database

        Args:
            history_minutes: How many history minutes to use
        """

        super().__init__()

        self.history_minutes = history_minutes
        self.history_duration = pd.Timedelta(self.history_minutes, unit="minutes")

    def __iter__(self):
        df_metadata = get_metadata_from_pvsites_database()
        df_gen = get_pv_power_from_pvsites_database(history_duration=self.history_duration)

        # Database record is very short. Set observed max to NaN
        df_metadata["observed_capacity_wp"] = np.nan

        # Ensure systems are consistant between generation data, and metadata
        common_systems = list(np.intersect1d(df_metadata.index, df_gen.columns))
        df_gen = df_gen[common_systems]
        df_metadata = df_metadata.loc[common_systems]

        # Compile data into an xarray DataArray
        xr_array = put_pv_data_into_an_xr_dataarray(
            df_gen=df_gen,
            observed_system_capacities=df_metadata.observed_capacity_wp,
            nominal_system_capacities=df_metadata.nominal_capacity_wp,
            ml_id=df_metadata.ml_id,
            latitude=df_metadata.latitude,
            longitude=df_metadata.longitude,
            tilt=df_metadata.get("tilt"),
            orientation=df_metadata.get("orientation"),
        )

        logger.info(f"Found {len(xr_array.ml_id)} PV systems")

        while True:
            yield xr_array


def get_metadata_from_pvsites_database() -> pd.DataFrame:
    """Load metadata from the pvsites database"""
    # make database connection
    url = os.getenv("DB_URL_PV")
    db_connection = DatabaseConnection(url=url, base=Base_Forecast)

    with db_connection.engine.connect() as conn:
        df_sites_metadata = pd.DataFrame(conn.execute(text("SELECT * FROM sites")).fetchall())

    df_sites_metadata["nominal_capacity_wp"] = df_sites_metadata["capacity_kw"] * 1000

    df_sites_metadata = df_sites_metadata.set_index("site_uuid")

    return df_sites_metadata


def get_pv_power_from_pvsites_database(history_duration: timedelta):
    """Load recent generation data from the pvsites database"""

    # make database connection
    url = os.getenv("DB_URL_PV")
    db_connection = DatabaseConnection(url=url, base=Base_Forecast)

    columns = "site_uuid, generation_power_kw, start_utc, end_utc"

    start_time = f"{datetime.now() - history_duration}"

    with db_connection.engine.connect() as conn:
        df_db_raw = pd.DataFrame(
            conn.execute(
                text(f"SELECT {columns} FROM generation where end_utc >= '{start_time}'")
            ).fetchall()
        )

    # Reshape
    df_gen = df_db_raw.pivot(index="end_utc", columns="site_uuid", values="generation_power_kw")

    # Rescale from kW to W
    df_gen = df_gen * 1000

    # Fix data types
    df_gen = df_gen.astype(np.float32)

    return df_gen



""" Make fake GSP data """

from datetime import datetime

import numpy as np

from ocf_datapipes.batch import BatchKey
from ocf_datapipes.batch.fake.utils import get_n_time_steps_from_config, make_time_utc
from ocf_datapipes.config.model import Configuration


def make_fake_gsp_data(
    configuration: Configuration, t0_datetime_utc: datetime, batch_size: int = 8
) -> dict:
    """
    Make Fake GSP data ready for ML model

    Args:
        configuration: configuration object
        t0_datetime_utc: one datetime for when t0 is
        batch_size: Integer batch size to create

    Returns: dictionary of gsp items
    """

    gsp_config = configuration.input_data.gsp

    if gsp_config is None:
        return {}

    n_gsps = gsp_config.n_gsp_per_example
    n_fourier_features = 8

    time_utc = make_time_utc(
        batch_size=batch_size,
        history_minutes=gsp_config.history_minutes,
        forecast_minutes=gsp_config.forecast_minutes,
        t0_datetime_utc=t0_datetime_utc,
        time_resolution_minutes=gsp_config.time_resolution_minutes,
    )
    n_times = time_utc.shape[1]

    batch = {}
    batch[BatchKey.gsp_time_utc] = time_utc  # Seconds since UNIX epoch (1970-01-01).
    batch[BatchKey.gsp] = np.random.random((batch_size, n_times, n_gsps))
    batch[BatchKey.gsp_nominal_capacity_mwp] = np.random.randint(
        0, 1000, (batch_size, n_times, n_gsps)
    )
    batch[BatchKey.gsp_effective_capacity_mwp] = np.random.randint(
        0, 1000, (batch_size, n_times, n_gsps)
    )
    batch[BatchKey.gsp_t0_idx] = get_n_time_steps_from_config(
        input_data_configuration=gsp_config, include_forecast=False
    )
    batch[BatchKey.gsp_id] = np.random.randint(0, 1000, (batch_size, n_gsps))
    batch[BatchKey.gsp_x_osgb] = np.random.randint(0, 10**6, (batch_size, n_gsps))
    batch[BatchKey.gsp_y_osgb] = np.random.randint(0, 10**6, (batch_size, n_gsps))

    batch[BatchKey.gsp_x_osgb_fourier] = np.random.random((batch_size, n_gsps, n_fourier_features))
    batch[BatchKey.gsp_y_osgb_fourier] = np.random.random((batch_size, n_gsps, n_fourier_features))
    batch[BatchKey.gsp_time_utc_fourier] = np.random.random(
        (batch_size, n_times, n_fourier_features)
    )

    return batch
